import __builtin__
import errno
import os
import sys
import string
import thread
import threading
import time

from ErrorFormatter import ErrorFormatter

# the identity of the main thread
MAIN_THREAD = thread.get_ident()


__ALPHA36 =  string.digits + string.letters


def _pretty_excepthook(mytype, value, traceb):

    out = "=== Unhandled error! Something bad and unexpected happened.\n"
    out += ErrorFormatter().format(tb = (mytype, value, traceb),
                                   deep_trace = True)
    log(out)

# this is not really useful at the moment; thus it's disabled
sys.excepthook = _pretty_excepthook


#
# Convenient logger.
#
def _log(data, is_warning = False):

    # show warnings only once
    if (is_warning):
        if (data in _warnings_so_far): return
        else: _warnings_so_far[data] = 1


    now = time.strftime("%D-%T", time.localtime(time.time()))
    out = "[" + now + "]==="
    out = "=" * (80 - len(out)) + out + "\n"
    out += data + "\n"
    out = "\n" + out + "\n"

    print >> sys.stderr, out


# set up logger
_warnings_so_far = {}
__builtin__.log = _log


def parse_color(color):

    if (color[0] == "#" and len(color) == 9):
        alpha = int(color[-2:], 16)
        color = color[:-2]
    else:
        alpha = 255

    import gtk
    c = gtk.gdk.color_parse(color)
    return (c.red >> 8, c.green >> 8, c.blue >> 8, alpha)


def radix(n, b):

    """
    Inverse function to int/long.

    int(radix(n, b), b) == n

    @param n: number to print
    @type n: int/long

    @param b: base
    @type b: 2 <= b <= 36

    @return: str representing n in base b
    @rtype: str
    """

    if not 2 <= b <= 36:
        raise ValueError, "base must be in [2; 36]"

    if n == 0:
        return 0

    if n < 0:
        sign = "-"
        n = -n
    else:
        sign = ""

    rep = ""

    while n:
        n, r = divmod(n, b)
        rep += __ALPHA36[r]

    return sign + rep[::-1]


def makedirs(path):

    """Recursive directory creation function

    Returns :
    - True if path has been created
    - False if path already exists

    (Re)Raises an OSError exception if path does not exist and
    cannot be created or if path cannot be accessed"""

    try:
        os.makedirs(path)
        return True

    except OSError, e:

        if e.errno != errno.EEXIST or not os.path.isdir(path):
            raise

        if not os.access(path, os.F_OK | os.R_OK | os.W_OK | os.X_OK):
            raise OSError("[Errno %d] %s: '%s'"
                          %
                          (errno.EACCES,
                           os.strerror(errno.EACCES),
                           path)
                          )

        return False


def _request_call(when_idle, function, *args):
    """
    Runs the given function by a timer. If several calls of that function
    occured before, only the last one is run. Do not call this function
    directly, instead use one of its wrappers 'request_call' and
    'request_idle_call'.
    """
    assert (function != None)

    import gobject

    def f(tstamp, function, args):
        if (_timestamps[function] > tstamp): return
        else: function(*args)

    tstamp = time.time()
    _timestamps[function] = tstamp
    if (when_idle):
        gobject.idle_add(f, tstamp, function, args)
    else:
        gobject.timeout_add(0, f, tstamp, function, args)


def request_call(function, *args): _request_call(False, function, *args)
def request_idle_call(function, *args): _request_call(True, function, *args)

        
_timestamps = {}


def run_in_main_thread(function, *args):

    """
    Runs the given function in the main thread.

    @param function:
    @type  function: callable

    @param *args: arguments
    """

    import gobject

    if (thread.get_ident() == MAIN_THREAD):
        try:
            return function(*args)
        except Exception, exc:
            log("Warning: %s(%s) has raised an exception while running in "
                "main thread.\nThe error was:\n%s" % (function, args, exc))
            import traceback; traceback.print_exc()
            return

    else:
        event = threading.Event()
        result = [None]

        def tmout():
            try:
                result[0] = function(*args)
            except Exception, exc:
                log("Warning: %s(%s) has raised an exception while running in "
                    "main thread.\nThe error was:\n%s" % (function, args, exc))

            event.set()

        gobject.timeout_add(0, tmout)

        event.wait()
        return result[0]


def run_nonblocking(function, *args):
    """
      Runs the given function in a new thread while the main thread waits in
      a nonblocking way, i.e. runs the GTK mainloop.
      If the given function throws an exception, that exception is rethrown by
      this function so that it doesn't get lost.
    """

    def worker(function, event, retval, *args):

        success = False
        try:
            ret = function(*args)
            success = True
        except:
            ret = sys.exc_value
            success = False

        retval[0] = (success, ret)
        event.set()


    retval = [None]
    event = threading.Event()
    t = threading.Thread(target = worker,
                         args = [function, event, retval] + list(args))
    t.start()

    import gtk
    while (not event.isSet()):
        gtk.threads_enter()
        gtk.mainiteration()
        gtk.threads_leave()
    success, ret = retval[0]

    if (not success):
        raise ret
    else:
        return ret


#
# Bind binds a function with arguments
# e.g.  f_with_arg = Bind(f, arg)  ->  f_with_0() <-> f(arg)
#
#
# Memento on *args
# the * notation expands its right_value
# sum4( *(1, 2, 3, 4) ) <-> sum4(1, 2, 3, 4))
#
#
# Bind also works with n-ary functions (including *args)
#
# >>> def sum4(a, b, c, d): return a + b + c +d
# >>> sum2 = Bind(sum4, 1, 2)
# >>> sum2(3, 4)
# 10
# >>> sum4(1, 2, 3, 4)
# 10
#
# Bind makes you able to call a function with some fixed arguments, without
# writing them. It can help you to shadow something.
# "string".lower() is in fact a str instance bound to strobject.lower
# "string".lower() <-> str.lower("string")
#
#
# Note on implementation :
# Bind is implemented with lambda expression because they are faster than
# functions.
# We want to call bound_f = Bind(f, bound_arg0, bound_args1, ...)
# so the first lambda expression at the upper level takes two arguments
# f and its arguments (as an arguments list so it can fit everywhere)
# What we want is to hide bound_args, so we have to return a lambda
# i.e a callable object that only takes arguments.
# At last, the body of the expression that will be evaluated it simple :
# We want to call f with its bound_args and with additional args
# bound_args and args are both tuples, so we just concatenate them (+)
# then we restore the call semantic (*)

# Bind = lambda f, *bound_args : lambda *args : f(* (bound_args + args) )

def Bind(f, *bound_args):

    def helper(*args):
        return f(* (bound_args + args) )

    return helper

#
# The idea behind FakeSelf is to provide member function call syntax for
# non-member function.
# self.attribute = FakeSelf(f, arg)
# self.attribute(extra) <-> f(arg, extra)
# so we have to remove the first argument self, in order to call f.
#
# FakeSelf also provides binding
#
# Note on implementation:
# FakeSelf looks like Bind except that the inner lambda has a first extra argument
# self, that is ignored.
#
# NB:
# FakeSelf can be implemented using Bind
# FakeSelf = lambda f, *bound_args : Bind( lambda self, *args : f(*args), *bound_args)


# FakeSelf = lambda f, *bound_args : lambda self, *args : f(* (bound_args + args) )

def FakeSelf(f, *bound_args):

    def helper(self, *args):
        return f(* (bound_args + args) )

    return helper




from DisplayTarget import DisplayTarget
from utils import dialog
from utils.datatypes import *
from layout import Unit
import utils

import gobject
import gtk


#
# Abstract class for targets that can contain others.
#
class ContainerTarget(DisplayTarget):

    __slots__ = ('__children', '__children_id', '__ids', '__relatives')

    def __init__(self, name, parent):

        # the list is necessary for preserving order
        self.__children = []

        # mapping: id -> child
        self.__children_id = {}

        # mapping: child -> id
        self.__ids = {}

        # the relative-relations: id_of_relative -> [child]
        self.__relatives = {}

        DisplayTarget.__init__(self, name, parent)



    #
    # Observer callback for watching children.
    #
    def child_observer(self, src, cmd): pass



    #
    # Creates a new child of the given type.
    #
    def new_child(self, childtype, settings, children):

        import targetregistry
        cid = settings["id"]

        child = targetregistry.create(childtype, self)
        only_one = targetregistry.one_child(self.get_name())

        if (only_one and len(self.__children) > 0):
            log("Warning: The %s container accepts only one child."
                % self.get_name(),
                is_warning = True)
        self._register_child(child, cid)
        child.get_widget().show()

        for t, s, c in children:
            child.new_child(t, s, c)

        for key, value in settings.items():
            child.set_xml_prop(key, value)

        return child



    def _register_child(self, child, cid):

        self.__children_id[cid] = child
        self.__ids[child] = cid
        self.__children.append(child)
        self._get_display().add_target_to_script(cid, child)



    def _unregister_child(self, child):

        cid = self.__ids[child]

        try:
            del self.__children_id[cid]
        except KeyError:
            pass

        try:
            del self.__ids[child]
        except KeyError:
            pass

        self.__children.remove(child)



    def _get_children(self): return self.__children[:]



    def _get_child(self):

        if (not self.__children): return
        else: return self.__children[0]



    def get_child_by_id(self, ident): return self.__children_id.get(ident)
    def get_id_by_child(self, child): return self.__ids.get(child)



    def delete(self):

        DisplayTarget.delete(self)
        for c in self._get_children():
            c.delete()
            self._unregister_child(c)
            del c
        self.__children = []
        self.__children_id.clear()
        self.__ids.clear()



    #
    # Override this method if the container size differs from the target size.
    #
    def get_container_geometry(self):

        return self.get_geometry()



    #
    # Returns the border size of this container. Override this method to return
    # the sizes of the four borders (left, top, right, bottom).
    #
    def get_border_size(self):

        return (Unit.ZERO, Unit.ZERO, Unit.ZERO, Unit.ZERO)




    #
    # Containers which set index values for the children override this method.
    #
    def get_next_child_index(self): return -1




    def detect_leave(self, stamp):

        if (not self._is_active()): return
        DisplayTarget.detect_leave(self, stamp)
        for c in self.__children:
            c.detect_leave(stamp)


    def unlock_geometry(self):

        self._geometry_lock = False
        for c in self._get_children(): c.unlock_geometry()




from Vault import Vault
from main.Control import Control
from plugin.Interface import Interface


#
# Wrapper for controls.
#
class ControlWrapper(object):

    def __init__(self, control):

        self.__dict__["_ControlWrapper__control"] = Vault(control)
        self.__dict__["_ControlWrapper__properties"] = \
                     Vault( dict(Interface.get_properties(control.__class__)) )

        ids =  [ Interface.get_id(i)
                 for i in Interface.get_interfaces( control.__class__ ) ]
        taz_ids = [ Interface.get_taz_style_id(i)
                    for i in Interface.get_interfaces( control.__class__ ) ]

        self.__dict__["_ControlWrapper__ifaces_id"] = \
                                                  Vault( tuple(ids + taz_ids) )



    def __setattr__(self, name, value):

        try:
            prop = self.__properties(open)[name]
        except KeyError:
            log("Warning: Property \"%s\" isn't available." % (name,))
            return

        prop.fset(self.__control(open), value)



    def __getattr__(self, name):

        if name in Control.AUTHORIZED_METHODS:
            return getattr(self.__control(open), name)

        try:
            prop = self.__properties(open)[name]
        except KeyError:
            log("Warning: Property \"%s\" isn't available." % (name,))
            return Null

        return prop.fget(self.__control(open))



    def get_interfaces_id(self):

        """
        @return : implemented interfaces' id
        @rtype : list of str
        """

        return self.__ifaces_id(open)




from shell.ListView import ListView

import gtk
import gobject


class BookmarkList(gtk.VBox):

    def __init__(self):

        self.__entries = []
        self.__categories = {}
        self.__category_names = []

        gtk.VBox.__init__(self)
        self.__listview = ListView((gtk.gdk.Pixbuf, gobject.TYPE_STRING),
                                   self.__item_renderer)
        self.__listview.connect("cursor-changed", self.__on_select_item)

        self.__optmenu = gtk.combo_box_new_text()
        self.__optmenu.connect("changed", self.__on_select_category)
        self.pack_start(self.__optmenu, False, False, 4)

        scrwin = gtk.ScrolledWindow()
        scrwin.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrwin.add(self.__listview)
        self.pack_end(scrwin, True, True, 0)
        
        self.show_all()


    def select_category(self, n):

        self.__optmenu.set_active(n)
        


    def add_category(self, name, label):

        if (name in self.__categories):
            self.clear_category(name)
        else:
            self.__optmenu.append_text(name)
            self.__categories[name] = []
            self.__category_names.append(name)
        


    def add_item(self, category, icon, label, callback, *args):
        assert (category in self.__categories)

        self.__categories[category].append((icon, label, callback, args))


    def clear_category(self, category):
        assert (category in self.__categories)

        self.__categories[category] = []


    def __on_select_item(self, src):

        path, col = src.get_cursor()
        callback, args = self.__entries[path[0]]
        if (callback): callback(*args)


    def __on_select_category(self, src):

        index = self.__optmenu.get_active()
        name = self.__category_names[index]
        items = self.__categories[name]
        self.__listview.clear()
        self.__entries = []
        for i in items:
            icon, label, callback, args = i
            self.__listview.add_item((icon, label))
            self.__entries.append((callback, args))
        #end for

    def __item_renderer(self, item):

        icon, label = item
        label = label.replace("<" ,"&lt;")
        return (icon, label)



from StateSaver import DefaultStateSaver
from ConfigDialog import ConfigDialog
from main import ICON, VERSION
from utils.datatypes import *
import settings

import getopt
import sys


class DaemonConfigger(ConfigDialog):
    """
      Configuration Dialog for the daemon.
    """

    __ITEMS = (
      ("title", {"label": "Editor to view/edit the desklet source code"}),
      ("uri",   {"label": "Your favorite editor", "bind": "editor"}),
      ("title", {"label": "XComposite support (requires composition manager)"}),
      ("boolean", {"label": "Translucency (takes effect after restarting a display)", "bind": "translucency"}),
      ("title", {"label": "Screen Resolution (DPI)"}),
      ("dpi",   {"label": "<small>" +
                          _("Adjust the value above so that the bar will be "
                            "exactly <b>5 cm</b> or <b>1.97\"</b> wide") +
                          "</small>",
                          "bind": "dpi", "value": 96})
      )


    def __init__(self):

        self.__backend = DefaultStateSaver()

        ConfigDialog.__init__(self)

        self.set_property("title", _("Configuration"))
        self.set_banner(ICON, "<big>gDesklets Configuration</big>\n"
                        "Version %s" % (VERSION,))

        self._set_setter(self.__setter)
        self._set_getter(self.__getter)
        self._set_caller(self.__caller)

        self.build(self.__ITEMS)

        self.__load_config()
        self.__read_cmd_line()



    def __setter(self, key, value, datatype):

        if (key == "editor"):
            settings.editor = value

        elif (key == "dpi"):
            settings.dpi = value

        elif (key == "translucency"):
            settings.translucent = value

        self.__backend.set_key(key, value)


    def __getter(self, key):

        if (key == "editor"):
            return settings.editor

        if (key == "dpi"):
            return settings.dpi

        elif (key == "translucency"):
            return settings.translucent

        else:
            return "gDesklets killed a kitten!"


    def __caller(self, *args): pass


    def __load_config(self):

        settings.editor = self.__backend.get_key("editor", settings.editor)
        settings.dpi = self.__backend.get_key("dpi", settings.dpi)
        settings.translucent = self.__backend.get_key("translucency", settings.translucent)


    def __read_cmd_line(self):

        OPTIONS = ("sm-client-id=", "sm-config-prefix=", "sm-disable",
                   "no-tray-icon", "translucent")

        #
        # Parses the given list of command line arguments. This is usually
        # sys.argv[1:].
        #

        try:
            opts, rest = getopt.getopt(sys.argv[1:], "nop:v", OPTIONS)
        except getopt.GetoptError:
            return

        for o, a in opts:
            if (o == "--translucent"):
                settings.translucent = True
            elif (o == "--no-tray-icon"):
                settings.show_tray_icon = False
            elif (o in ("--sm-client-id", "--sm-config-prefix",
                        "--sm-disable")):
                pass
        #end for



from shell.Plugin import Plugin

import gtk


#
# The switch buttons for the different views.
# Reads the available views from the "views" file.
#
class Shell_ViewSwitch(Plugin):

    def init(self):
        self.__buttons = []
        self.__views = []

        vbox = gtk.VBox(True, 4)
        vbox.set_border_width(4)

        views = self._get_plugins_by_pattern("name", "View_*")

        cnt = 0
        views.sort(lambda x, y: cmp(x.get_priority(), y.get_priority()))
        for view in views:
            #viewname = v.strip()
            #if (not viewname): continue
            
            if (cnt == 0):
                hbox = gtk.HBox(True, 4)
                vbox.add(hbox)

            #view = self._get_plugin(viewname)
            btn = gtk.ToggleButton()
            align = gtk.Alignment(0.5, 0.5)
            btn.add(align)
            bbox = gtk.HBox(False, 2)
            bbox.add(view.get_icon())
            bbox.add(gtk.Label(view.get_name()))
            align.add(bbox)
            
            btn.connect("toggled", self.__on_switch, view)
            self.__buttons.append(btn)
            self.__views.append(view)
            hbox.add(btn)
            cnt += 1
            if (cnt == 2): cnt = 0
        #end for

        self.__buttons[0].set_active(True)
        vbox.show_all()
        self.__widget = vbox

        shell = self._get_plugin("UI_Shell")
        shell.set_switch(self.__widget)



    def __on_switch(self, src, view):

        if (src.get_active()):
            for btn in self.__buttons:
                if (btn != src): btn.set_active(False)

            for v in self.__views:
                if (v != view): v.deactivate()

            view.activate()


def get_class(): return Shell_ViewSwitch



from ConfigFloat import ConfigFloat


class ConfigInteger(ConfigFloat):

    def __init__(self, name, getter, setter, caller):

        ConfigFloat.__init__(self, name, getter, setter, caller,
                             int_only = True)



import os
import time

from utils import vfs


#
# Class for storing the list of profiles and their displays.
#
class DisplayList:

    __ADD = 0
    __DEL = 1

    __LOCK = "/var/lock/gdesklets.lock"

    def __init__(self, displaylist):

        # the path of the display list file
        self.__display_list = displaylist

        # the current profile
        self.__current_profile = "default"

        # table for profile -> [id]
        self.__profiles = {}

        # table for id -> path
        self.__path_table = {}

        # table for id -> profile
        self.__profiles_table = {}

        # the list of unsaved modifications to the list
        self.__modifications = []

        self.__load_list()



    #
    # Loads the display list from file.
    #
    def __load_list(self):

        self.__profiles = {}
        self.__path_table = {}
        self.__profiles_table = {}

        try:
            data = open(self.__display_list, "r").readlines()
        except IOError, exc:
            return

        try:
            self.__current_profile = data.pop(0).strip()
        except IndexError, exc:
            log("The displaylist file has wrong format, ignoring it.")
            return

        self.__profiles[self.__current_profile] = []

        for line in data:
            if (not line): continue

            try:
                ident, path, profile = line.split()
            except ValueError, exc:
                log("The displaylist file has wrong format, ignoring it.")
                return

            path = vfs.unescape_path(path)
            profile = vfs.unescape_path(profile)
            if (not profile in self.__profiles):
                self.__profiles[profile] = []
            self.__profiles[profile].append(ident)

            self.__path_table[ident] = path
            self.__profiles_table[ident] = profile



    #
    # Saves the display list to file.
    #
    def __save_list(self):

        data = [self.__current_profile + "\n"]
        for profile, idents in self.__profiles.items():
            for ident in idents:
                path = self.__path_table[ident]
                path = vfs.escape_path(path)
                profile = vfs.escape_path(profile)
                data.append("%s %s %s\n" % (ident, path, profile))

        try:
            open(self.__display_list, "w").writelines(data)
        except IOError, exc:
            log("Could not open file \"%s\" for writing data!\n%s\n"
                % (self.__display_list, exc))



    #
    # Returns whether there are unsaved modifications.
    #
    def __is_dirty(self):

        return bool(self.__modifications)



    #
    # Returns the current profile.
    #
    def get_profile(self):

        return self.__current_profile



    #
    # Sets the current profile.
    #
    def set_profile(self, profile):

        self.__current_profile = profile



    #
    # Returns a list of available profiles.
    #
    def get_profiles(self):

        if (self.__is_dirty()):
            self.commit()
        self.__load_list()

        return self.__profiles.keys()



    #
    # Returns a list of all displays in the given profile.
    #
    def get_displays(self, profile):

        if (self.__is_dirty()):
            self.commit()
        self.__load_list()

        return self.__profiles.get(profile, [])



    #
    # Looks up the data for the given display.
    #
    def lookup_display(self, ident):

        try:
            profile = self.__profiles_table[ident]
            path = self.__path_table[ident]

        except KeyError:
            raise KeyError("Could not find the path or the profile for ID "
                           "\"%s\"!" % ident)

        return (profile, path)



    #
    # Adds a new display.
    #
    def add_display(self, profile, path, ident):

        if (ident):
            self.__modifications.append((self.__ADD, profile, path, ident))



    #
    # Removes a display.
    #
    def remove_display(self, ident):

        try:
            profile, path = self.lookup_display(ident)
        except KeyError, exc:
            log("The following error occured while removing display \"%s\"!"
                 % exc)
            return

        self.__modifications.append((self.__DEL, profile, path, ident))



    #
    # Commits all changes. You have to call this method after making changes to
    # the list.
    #
    def commit(self):

        while (os.path.exists(self.__LOCK)):
            time.sleep(0.1)

        try:
            lockf = file(self.__LOCK, "w")
            lockf.close()
        except IOError:
            pass

        profile = self.__current_profile
        self.__load_list()
        self.__current_profile = profile

        for cmd, profile, path, ident in self.__modifications:
            if (cmd == self.__ADD and not ident in self.__path_table):
                if (not profile in self.__profiles):
                    self.__profiles[profile] = []
                self.__profiles[profile].append(ident)
                self.__path_table[ident] = path
                self.__profiles_table[ident] = profile

            elif (cmd == self.__DEL):
                del self.__path_table[ident]
                del self.__profiles_table[ident]
                self.__profiles[profile].remove(ident)

        self.__modifications = []
        self.__save_list()

        try:
            os.remove(self.__LOCK)
        except OSError:
            pass



from libdesklets.system.Arch import Arch

import re
import struct


class Generic(Arch):

    def __init__(self):

        Arch.__init__(self)

        self.__net_devices = re.compile('^\s*(\w+):.*mtu', re.M
                                        ).findall( os.popen('/sbin/ifconfig -a').read() )

        # just get the first cpu0 line

        # cpu0: Intel Celeron (Mendocino) ("GenuineIntel" 686-class, 128KB L2 cache) 468 MHz
        # cpu0: FPU,V86,DE,PSE,TSC,MSR,PAE,MCE,CX8,SEP,MTRR,PGE,MCA,CMOV,PAT,PSE36,MMX,FXSR

        m = re.search('^cpu0: (.*?) \(.*?\)* (\d+) MHz$',
                      open("/var/run/dmesg.boot").read(),
                      re.MULTILINE)

        self.__model = m.group(1)
        self.__speed = m.group(2)



    def net_devices(self):
        """
        @return : all available network devices
        @rtype  : list
        """

        return (self.__net_devices)



    def cpu_model(self):
        """
        @return : model/type of installed processor
        @rtype  : str
        """

        return (self.__model)



    def cpu_speed(self):
        """
        @return : current clock of installed processor
        @rtype  : float
        """

        return float(self.__speed)



    def users(self):
        """
        @return : number of connected users
        @rtype  : int
        """

        # man utmp
        # don't know if
        # sizeof(struct utmp) == 300
        # offsetof(struct utmp, ut_name) == 8
        # on every OpenBSD arch
        # sparc64 : ok

        count = 0
        data = open('/var/run/utmp', 'rb').read()

        for i in range(8, len(data), 300):

            ut_name = struct.unpack('c', data[i:i+1])[0]

            if (ut_name != '\0'):
                count += 1

        return count



from utils.Observable import Observable

import gtk


class SearchBar(gtk.HBox, Observable):

    OBS_UPDATE = 0
    

    def __init__(self):

        gtk.HBox.__init__(self, False, 4)
        self.set_border_width(4)

        self.__query = gtk.Entry()
        self.__query.connect("key-press-event", self.__on_key)
        self.pack_start(self.__query, True, True, 0)

        searchbtn = gtk.Button(_("Search"))
        searchbtn.connect("clicked", self.__on_click, self.__query)
        self.pack_end(searchbtn, False, False, 0)



    def set_query(self, query):

        self.__query.set_text(query)
        self.update_observer(self.OBS_UPDATE, query)
        

    def __on_key(self, src, event):

        key = gtk.gdk.keyval_name(event.keyval)
        if (key in ("Return", "KP_Enter")):
            self.update_observer(self.OBS_UPDATE, src.get_text())



    def __on_click(self, src, query):

        self.update_observer(self.OBS_UPDATE, query.get_text())



from utils.Element import Element
from utils.Observable import Observable
from utils.datatypes import *
from utils import actionparser


#
# Abstract class for data targets.
#
class DataTarget(Element, Observable):

    __slots__ = ('__actions', '__index_path', '__parent', '__display',
    '__watch_bindings', '__index')

    AUTHORIZED_METHODS = ()

    def __init__(self, name, parent):

        # the actions and their associated calls
        self.__actions = {}

        # the index path of this target
        self.__index_path = []

        # the parent of this target
        self.__parent = parent

        # the display of this target
        self.__display = parent._get_display()

        # list of watch bindings for being able to unbind the sensor again
        # FIXME: remove eventually :)
        self.__watch_bindings = []

        Element.__init__(self, name)

        self._register_property("event", TYPE_OBJECT, None, self._getp)
        self._register_property("index-path", TYPE_INT, None, self._getp)
        self._register_property("watch", TYPE_LIST, self._setp_watch, None)

        if (parent and not self.is_standalone()):
            self.__index = parent.get_next_child_index()
            self.__index_path = parent.get_index_path()
        else:
            self.__index = -1
            self.__index_path = []

        if (not self.is_standalone()):
            self.add_observer(parent.child_observer)

        if (self.__index != -1): self.__index_path.append(self.__index)
        if (self.__index_path):
            self._setp("index-path", self.__index_path[:])
        else:
            self._setp("index", [])



    #
    # Returns the parent of this target.
    #
    def _get_parent(self): return self.__parent



    #
    # Returns whether this target is standalone, i.e. needs no parent.
    #
    def is_standalone(self): return True



    #
    # Tells this target about its XML name.
    #
    def set_name(self, name): self.__name = name


    def delete(self):

        self.drop_observers()
        self.__unbind_sensors()



    #
    # Returns the display of this target.
    #
    def _get_display(self): return self.__display



    #
    # Returns the sensor watch propagator of this target. This is either the
    # display window or an array.
    #
    def _get_propagator(self):

        parent = self._get_parent()
        while (parent and not parent.__class__.__name__ == "TargetArray"):
            parent = parent._get_parent()

        if (not parent): parent = self._get_display()

        return parent



    #
    # Returns the index path of this target.
    #
    def get_index_path(self): return self.__index_path[:]



    #
    # Registers the given action.
    #
    def _register_action(self, action):

        self._register_property("on-" + action, TYPE_STRING,
                                self._setp__action, self._getp)



    #
    # Sets the given XML property. This is the same as set_prop() but takes
    # strings as values and converts before setting.
    #
    def set_xml_prop(self, key, value):

        self.set_prop_from_string(key, value)



    #
    # Emits the given action if the pointer is over this target.
    #
    def get_action_call(self, action):

        return self.__actions.get(action, None)



    #
    # Handles the given action.
    #
    def handle_action(self, action, px, py, event):

        call = self.get_action_call(action)
        if (call):
            # analyze call to see if it's a legacy call
            # FIXME: remove eventually :)
            import re
            if (re.match("[\w\-]+:.*", call)):
                log("Deprecation: Please use new style call.",
                    is_warning = True)

                try:
                    legacy_args = event._args
                except:
                    legacy_args = []

                legacy_call = actionparser.parse(call)
                path = self.get_index_path()
                self._get_display().call_sensor(legacy_call, path,
                                                *legacy_args)

            else:
                self._setp("event", event)
                self._get_display().execute_callback_script(call, self)


    #
    # Unbinds this target from the sensors.
    # FIXME: remove eventually :)
    #
    def __unbind_sensors(self):

        for sensorplug, prop in self.__watch_bindings:
            self._get_display().unbind_sensor(sensorplug, self, prop)
        #end for
        self.__watch_bindings = []



    #
    # Action handlers.
    #
    def _setp__action(self, key, value):

        name = key[3:]
        self.__actions[name] = value
        self._setp(key, value)



    #
    # "watch" property.
    #
    def _setp_watch(self, key, value):

        entries = value
        for e in entries:
            prop, sensorplug = e.split("=")
            if (self.__index_path):
                self._get_display().register_array_for_port(
                    self._get_propagator(),
                    sensorplug)
                sensorplug += str(self.__index_path)
            self._get_display().bind_sensor(sensorplug, self, prop)
            self.__watch_bindings.append((sensorplug, prop))




import gtk

from plugin.Interface import Interface
from utils.HIGDialog import HIGDialog


#
# Dialog for inspecting controls.
#
class ControlInspector(HIGDialog):

    def __init__(self, ctrlclass):

        HIGDialog.__init__(self, buttons=(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
        name = ctrlclass.__name__
        self.set_property("title", _("%s Control") % (name))

        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.set_property("border-width", 6)
        self.__scrollw = gtk.ScrolledWindow()
        align.add(self.__scrollw)

        self.__label = gtk.Label()
        self.__label.set_property("selectable", True)

        self.vbox.pack_start(align, False, False, 0)
        self.vbox.pack_start(self.__label, False, False, 0)

        # list the interfaces and their descriptions
        texts = Interface.gui_describe(ctrlclass)

        # create the listview now, with the content
        self.__create_listview(texts)

        def f(*args): self.destroy()
        self.connect("response", f)

        self.vbox.show_all()


    def __create_listview(self, texts):

        self.__label.set_property("label", texts[0][0])

        import gobject, pango

        treemodel = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_STRING,
                                  gobject.TYPE_STRING, gobject.TYPE_STRING)
        treeview = gtk.TreeView(treemodel)

        textrenderer0 = gtk.CellRendererText()
        textrenderer0.set_property("weight", pango.WEIGHT_BOLD)
        textrenderer1 = gtk.CellRendererText()

        column0 = gtk.TreeViewColumn(_("Interface Name"), textrenderer0)
        column0.add_attribute(textrenderer0, "text", 0)
        treeview.append_column(column0)
        column1 = gtk.TreeViewColumn(_("Property Name"), textrenderer1)
        column1.add_attribute(textrenderer1, "text", 1)
        treeview.append_column(column1)
        column2 = gtk.TreeViewColumn(_("Property Access"), textrenderer1)
        column2.add_attribute(textrenderer1, "text", 2)
        treeview.append_column(column2)
        column3 = gtk.TreeViewColumn(_("Property Description"), textrenderer1)
        column3.add_attribute(textrenderer1, "text", 3)
        treeview.append_column(column3)

        for (iface, items) in texts:

            myiter = treemodel.insert_before(None, None,
                                             (iface.split(":")[0], None, None,
                                              None))

            for name, access, description in items:
                treemodel.insert_before(myiter, None,
                                        (None, name, access, description))

        treeview.expand_all()
        w, h = treeview.size_request()
        self.__scrollw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.__scrollw.set_size_request(max(w + 20, 600), min(h, 400))
        self.__scrollw.add(treeview)




from shell.Plugin import Plugin
from shell.ItemBrowser import ItemBrowser
from main import HOME
from utils import dialog

try:
    set
except NameError:
    from sets import Set as set

import gtk
import string
import gobject
import os
import textwrap



class UI_DisplayBrowser(Plugin):

    __PREVIEW_WIDTH = 48
    __NO_PREVIEW = os.path.join(HOME, "data", "gdesklets.png")


    def init(self):

        self.__location = ""
        self.__collection = self._get_plugin("Core_DisplayCollection")
        self.__collection.add_observer(self.__reload)
        self.__browser = ItemBrowser(self.__search_handler,
                                     self.__item_renderer,
                                     gtk.gdk.Pixbuf, gobject.TYPE_STRING)
        self.__browser.add_observer(self.__on_activate)

        self.__reload()

        menu = self._get_plugin("UI_Menu")

        menu.set_item("File/Slot", gtk.STOCK_DELETE,
                      _("R_emove selected desklet"), self.__on_remove)
        menu.set_separator("File/Slot")
        menu.set_item("File/Slot", gtk.STOCK_EXECUTE,
                      _("_Run (remote) desklet..."), self.__on_run)
        menu.set_item("File/Slot", gtk.STOCK_EXECUTE,
                      _("R_un selected desklet"), self.__on_activate)

        shell = self._get_plugin("UI_Shell")

        w1, w2 = self.__browser.get_widgets()
        shell.set_sidebar(w1)
        shell.set_body(w2)



    def show(self):

        w1, w2 = self.__browser.get_widgets()
        w1.show_all()
        w2.show_all()


    def hide(self):

        w1, w2 = self.__browser.get_widgets()
        w1.hide()
        w2.hide()


    def set_location(self, expr):

        self.__location = expr
        self.__browser.set_query(expr)


    def __reload(self):

        for name, items in \
           [(_("by category"), self.__get_categories(self.__collection)),
            (_("by author"), self.__get_authors(self.__collection)),
            (_("alphabetically"), self.__get_alphabetically(self.__collection))]:
            self.__browser.add_category(name, name, items)


    def __on_run(self, *args):

        def responder(src, response):
            text = src.vbox.get_children()[0].get_children()[-1].get_text()
            if (response == gtk.RESPONSE_ACCEPT):
                if (text):
                    client = self._get_plugin("Core_Client")
                    ident = client.open_display(text)
                else:
                    src.destroy()

        from utils.HIGDialog import HIGDialog
        dialog = HIGDialog(buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                      gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
        dialog.connect("response", responder)
        dialog.show()
        label = gtk.Label("Please enter a valid desklet URI here")
        label.show()
        entry = gtk.Entry()
        entry.show()

        vbox = dialog.vbox
        hbox = gtk.HBox()
        hbox.show()
        hbox.set_property("spacing", 12)
        hbox.set_property("border-width", 6)
        hbox.pack_start(label)
        hbox.pack_start(entry)
        vbox.pack_start(hbox)


    def __on_activate(self, *args):

        item = self.__browser.get_selected_item()
        if (item):
            client = self._get_plugin("Core_Client")
            path = os.path.abspath(item.path)
            ident = client.open_display(path)


    def __on_remove(self, *args):

        def remove(*args):
            try:
                self.__remove_display(item.path)
            except OSError:
                dialog.warning(_("Could not remove"),
                               _("The desklet file could not be removed. "
                                 "This most likely means that you do not have "
                                 "write permission on it."))
            else:
                self.__collection.remove(item)
                self.set_location(self.__location)


        item = self.__browser.get_selected_item()
        if (item):
            dialog.question(None,
                            _("Do you really want to remove this desklet ?"),
                            _("The desklet will be removed from your system "
                              "and no longer be available."),
                            (gtk.STOCK_CANCEL, Null),
                            (gtk.STOCK_DELETE, remove))



    def __search_handler(self, expr):

        status = self._get_plugin("Shell_StatusBar")
        try:
            matches = self.__collection.query(expr)
            status.set_status(_("Found %d desklets") % (len(matches)))
            matches.sort(lambda a,b: cmp(a.name, b.name))
        except:
            status.set_status(_("Invalid search expression."))
            matches = None

        self.__location = expr

        return matches



    def __item_renderer(self, item):

        path = item.path
        directory, filename = os.path.split(path)
        preview = item.preview.strip()
        if (preview): preview = os.path.join(directory, preview)
        name = item.name.strip() or os.path.splitext(filename)[0]
        version = item.version.strip() or "???"
        author = item.author.strip().replace("<", "&lt;") or "???"
        description = item.description.strip().replace("<", "&lt;") \
                      or _("No description available")
        description = "\n".join(textwrap.wrap(description, 66))

        try:
            pixbuf = gtk.gdk.pixbuf_new_from_file(preview or self.__NO_PREVIEW)
        except:
            pixbuf = gtk.gdk.pixbuf_new_from_file(self.__NO_PREVIEW)

        width, height = pixbuf.get_width(), pixbuf.get_height()
        scale = self.__PREVIEW_WIDTH / float(width)
        pixbuf = pixbuf.scale_simple(int(width * scale),
                                     int(height * scale),
                                     gtk.gdk.INTERP_BILINEAR)

        text = "<big><b>%(name)s</b></big> " \
               "<small>ver. %(version)s</small>\n" \
               "<small>%(path)s</small>\n" \
               "<i>by %(author)s</i>\n" \
               "%(description)s" % vars()

        return (pixbuf, text)



    def __get_categories(self, collection):

        icon = os.path.join(self._get_path(), "folder.png")
        values = []
        matches = collection.query("(MATCH 'category' '*')")
        categs = set([m.category for m in matches if m.category])
        for categ in categs:
            values += [(icon, c.strip(), "(MATCH 'category' '%s')" % c.strip())
                       for c in categ.split(",")]
        values += [(icon, "uncategorized", "(MATCH 'category' '')")]

        return values


    def __get_authors(self, collection):

        icon = os.path.join(self._get_path(), "author.png")
        values = []
        matches = collection.query("(MATCH 'author' '*')")
        authors = set([m.author for m in matches if m.author])
        for a in authors:
            values.append((icon, a, "(MATCH 'author' '%s')" % a))

        return values


    def __get_alphabetically(self, collection):

        icon = os.path.join(self._get_path(), "folder.png")
        values = []

        for c in string.ascii_uppercase:
            query = "(MATCH 'name' '%c*')" % (c,)
            if collection.query( query ):
                values.append((icon, c, query))

        return values


    #
    # Removes the given .display file. If there are no .display files left
    # after, the whole directory gets removed.
    # Raises an OSError if it fails.
    #
    def __remove_display(self, path):

        def find(p):
            files = os.listdir(p)
            for f in files:
                filepath = os.path.join(p, f)
                if (os.path.isdir(f)): found = find(filepath)
                else: found = False

                if (found or (os.path.isfile(filepath) and
                      os.path.splitext(f)[-1] == ".display")):
                    return True
            #end for
            return False

        dirpath = os.path.dirname(path)

        # remove the given file
        os.remove(path)

        # if there are no .display files left, remove the directory
        if (not find(dirpath)):
            os.system("rm -rf %s" % (dirpath))


def get_class(): return UI_DisplayBrowser



from ControlFactory import ControlFactory
from utils.TargetSettings import TargetSettings
from display.Display import Display
from utils.datatypes import *
from utils import dialog
from utils import vfs

from xml import sax

import os
import sys


_STATE_NONE = -1
_STATE_DISPLAY = 0
_STATE_META = 1
_STATE_SCRIPT = 2
_STATE_PREFS = 3
_STATE_SENSOR = 4  # deprecated
_STATE_CONTROL = 5

_SENSOR_CONTROL = "ISensor:2peqs3czo5mlrv5ibwwkn5dqc"


#
# Class for creating Displays from XML data.
#
class _DisplayFactory(sax.handler.ContentHandler):

    def __init__(self):

        # the current state of the parser
        self.__state = [_STATE_NONE]

        # a stack of (tagname, attributes) pairs for remembering the nesting
        # of elements
        self.__nesting_stack = []

        # a stack of (type, settings, children) tuples to remember the children
        # of elements; children is a list of (type, settings, children) tuples
        self.__children_stack = [[]]

        # a list of the sensors for the display
        self.__sensors = []

        # the config items of the display
        self.__config_items = []

        self.__path = ""
        self.__scripts = []
        self.__script = ""
        self.__id = ""
        self.__control_factory = ControlFactory()
        self.__prefs_callback = None
        self.__metadata = {}

        sax.handler.ContentHandler.__init__(self)



    def __reset(self):

        self.__scripts = []
        self.__config_items = []
        self.__children_stack = [[]]
        self.__nesting_stack = []
        self.__sensors = []
        self.__prefs_callback = None
        self.__metadata = {}


    #
    # Parses the given XML data and returns a new Display object.
    #
    def create_display(self, ident, data, dspfile):

        dsp = Display(ident, dspfile)

        self.__reset()
        self.__id = ident
        self.__path = os.path.dirname(dspfile)
        self.__filename = dspfile

        # parse display data
        try:
            sax.parseString(data, self)

        except sax._exceptions.SAXParseException, sax_exc:
            raise UserError(_("XML parse error"),
                            _("This .display file has invalid XML syntax."),
                            code = data, lineno = sax_exc.getLineNumber())

        # add the sensor controls
        for ident, sensorctrl in self.__sensors:
            dsp.add_sensor(ident, sensorctrl)

        # add the children and configure
        childtype, settings, children = self.__children_stack.pop()[0]
        dsp.set_content(childtype, settings, children)

        # set some meta data if available
        preview = self.__metadata.get("preview", "")
        name = self.__metadata.get("name", "")
        version = self.__metadata.get("version", "")
        author = self.__metadata.get("author", "")
        dsp.set_metadata(preview = preview,
                         name = name,
                         version = version,
                         author = author)
        dsp.build_configurator(self.__config_items)

        # add scriptlets
        for script, filename in self.__scripts:
            dsp.add_scriptlet(script, filename)

        self.__reset()

        return dsp



    #
    # Creates a TargetSettings object from the given Attributes object.
    #
    def __create_settings(self, attrs):

        settings = TargetSettings()
        for key, value in attrs.items():
            settings.set(key, value)

        return settings



    def startElement(self, name, attrs):

        attrs_dict = {}
        attrs_dict.update(attrs)

        if (name == "display"):
            self.__state.append(_STATE_DISPLAY)

        elif (name == "prefs"):
            # support a global prefs callback
            self.__prefs_callback = attrs.get("callback", None)
            self.__state.append(_STATE_PREFS)

        elif (name == "meta"):
            self.__state.append(_STATE_META)

        elif (name == "sensor"):
            log("Deprecation: Sensors are deprecated since v0.30.\n"
                "Please consider using controls and inline scripts.")
            self.__state.append(_STATE_SENSOR)

        elif (name == "control"):
            self.__state.append(_STATE_CONTROL)

        elif (name == "script"):
            self.__state.append(_STATE_SCRIPT)
            self.__script = ""

        if (self.__state[-1] == _STATE_DISPLAY):
            self.__children_stack.append([])

        # remember everything for later
        self.__nesting_stack.append((name, attrs_dict))



    def endElement(self, name):

        oname, attrs = self.__nesting_stack.pop()
        state = self.__state[-1]
        if (not name == oname):
            # nesting errors in XML are detected by the SAX parser; if we
            # get here, it means our parser is buggy, not the XML input
            log("Nesting error: expected %s, got %s." % (oname, name))
            return


        elif (state == _STATE_DISPLAY):
            settings = self.__create_settings(attrs)
            # if there is no ID given, guess a unique one
            ident = attrs.get("id", Display.make_id())
            settings["id"] = ident

            # build the tree of children from bottom up
            children = self.__children_stack.pop()
            self.__children_stack[-1].append((name, settings, children))

        elif (state == _STATE_META):
            self.__metadata = attrs

        elif (state == _STATE_SCRIPT):
            # Some distros ship a broken SAX parser and
            # their users are too lazy to reportbug. (doesn't have __contains__)
            # So we have to use has_key()
            # if ((name == "script") and ("uri" in attrs)):
            if (name == "script" and attrs.has_key("uri")):
                path = attrs["uri"]
                filename = os.path.join(self.__path, path)
                try:
                    self.__script += "\n" + vfs.read_entire_file(filename)
                except:
                    log("File doesn't exist or couldn't be loaded: %s"
                        % (path,))
                    dialog.warning(_("File doesn't exist or couldn't be "
                                     "loaded"),
                                   _("A file which contains a script isn't "
                                     "available for further execution."))
            else:
                filename = self.__filename

            # collect script by script to be able to execute them one by one
            # later; we would get problems with different indentation otherwise
            self.__scripts.append((self.__script, filename))

        # deprecated sensor stuff
        elif (state == _STATE_SENSOR):
            from utils import typeconverter
            ident = attrs["id"]
            moduledata = typeconverter.str2type(TYPE_LIST, attrs["module"])
            module = moduledata[0]
            args = moduledata[1:]
            sensorctrl = self.__control_factory.get_control(_SENSOR_CONTROL)

            if (sensorctrl):
                sensorctrl.sensor = (module, args)
                sensorctrl.config_id = self.__id + ident
                self.__sensors.append((ident, sensorctrl))
            else:
                raise RuntimeError(_("Could not load sensor"))

        elif (state == _STATE_CONTROL):
            ident = attrs["id"]
            interface = attrs["interface"]
            script = "%s = get_control('%s')" \
                     % (ident, interface)
            self.__scripts.append((script, self.__filename))

        elif (state == _STATE_PREFS):
            if (name == "prefs"):
                pass
            else:
                if (name in ("boolean", "color", "enum", "float", "font",
                             "integer", "string", "uri")
                    and not attrs.has_key("bind")):
                    dialog.warning(_("&lt;%s&gt; needs a bind attribute") % \
                                                                      (name,),
                                   _("The &lt;prefs&gt; section of this "
                                     "desklet file is broken."))
                    raise AttributeError("<%s> needs a bind attribute" % \
                                                                      (name,))

                # use global prefs callback if no callback was given
                if (self.__prefs_callback and not attrs.has_key("callback")):
                    attrs["callback"] = self.__prefs_callback

                self.__config_items.append((name, attrs))


        if (name in ("display", "sensor", "control", "meta",
                     "prefs", "script")):
            self.__state.pop()



    def characters(self, content):

        state = self.__state[-1]
        if (state == _STATE_SCRIPT):
            self.__script += str(content)


_singleton = _DisplayFactory()
def DisplayFactory(): return _singleton



from DisplayTarget import DisplayTarget
from utils.datatypes import *
from layout import Unit
from utils.tiling import Tiling
import utils

import gobject
import gtk
import pango
import re


# determine the locale
import commands

fail, CHARMAP = commands.getstatusoutput("locale charmap")

if (fail or not CHARMAP):
    log("Could not detect character encoding.")
    CHARMAP = "ASCII"


class TargetLabel(DisplayTarget):

    # regular expression for parsing font description strings
    __RE_FONT = re.compile("(?P<name>.+?)"
                           "(?P<size>[0-9\.]+)"
                           "(?P<unit>[a-z%]+)?$")


    def utf8_to_utf8(value):

        if (isinstance(value, unicode)):
            try:
                value = value.encode("UTF-8", "replace")
            except ValueError:
                log("Unicode -> UTF-8 convertion failed !!!")
                raise

        elif (not isinstance(value, str)):
            # int, float, long, list, tuple, instance
            value = str(value)

        return value


    def charmap_to_utf8(value):

        if (isinstance(value, str)):
            try:
                value = unicode(value, CHARMAP, "ignore")
            except LookupError:
                # may be it's already UTF-8 ?
                value = unicode(value, "UTF-8", "replace")

        if (isinstance(value, unicode)):
            try:
                value = value.encode("UTF-8", "replace")
            except ValueError:
                log("Unicode -> UTF-8 convertion failed !!!")
                raise

        else:
            # int, float, long, list, tuple, instance
            value = str(value)

        return value


    #
    # Converts the given string to UTF-8 format.
    #
    if (CHARMAP == 'UTF-8'):
        __utf8ify = staticmethod(utf8_to_utf8)
    else:
        __utf8ify = staticmethod(charmap_to_utf8)



    def __init__(self, name, parent):

        self.__old_value = ""
        self.__font_description = None
        self.__wrap_at = -1

        self.__size = (0, 0)

        DisplayTarget.__init__(self, name, parent)

        self.__widget = Tiling()
        self.__pango_context = self.__widget.get_pango_context()
        self.__pango_context.set_base_dir(pango.DIRECTION_LTR)
        self.__pango_layout = pango.Layout(self.__pango_context)
        self.__color = (0, 0, 0, 255)
        self.__font_description = self.__pango_context.get_font_description()
        self.__widget.show()

        # i guess a label has to accept everything printable for its value
        # these days
        self._register_property("value", TYPE_ANY,
                                self._setp_value, self._getp)
        self._register_property("color", TYPE_STRING,
                                self._setp_color, self._getp)
        self._register_property("font", TYPE_STRING,
                                self._setp_font, self._getp)
        self._register_property("wrap-at", TYPE_UNIT,
                                self._setp_wrap_at, self._getp)

        self._setp("value", "")
        self._setp("color", "black")
        self.set_prop("font", "Sans 8")
        self.set_prop("wrap-at", Unit.Unit(0, Unit.UNIT_PX))

        # watch the widget for geometry changes; we need this for percentual
        # font sizes
        self.add_observer(self.__on_observe_size)
        

    def get_widget(self): return self.__widget


    def __on_observe_size(self, src, cmd, *args):

        x, y, w, h = src.get_geometry()
        if (cmd == src.OBS_GEOMETRY):
            self.__size = (w.as_px(), h.as_px())
            self.__set_wrap(self.get_prop("wrap-at"))
            self.__set_font(self.get_prop("font"))
            self.__set_value(self.get_prop("value"))


    #
    # Renders the given text.
    #
    def __render_text(self, layout):

        width, height = layout.get_pixel_size()

        # render font
        pmap = gtk.gdk.Pixmap(gtk.gdk.get_default_root_window(),
                              width, height * 2)
        gc = pmap.new_gc()
        gc.set_foreground(self.__widget.get_colormap().alloc_color("black"))
        pmap.draw_rectangle(gc, True, 0, 0, width, height)
        gc.set_foreground(self.__widget.get_colormap().alloc_color("white"))
        pmap.draw_rectangle(gc, True, 0, height, width, height)

        r, g, b, a = self.__color
        col = "#" + ("0" + hex(r)[2:])[-2:] + \
                    ("0" + hex(g)[2:])[-2:] + \
                    ("0" + hex(b)[2:])[-2:]
        gc.set_foreground(
            gtk.gdk.get_default_root_window().get_colormap().alloc_color(col))
        pmap.draw_layout(gc, 0, 0, layout)
        pmap.draw_layout(gc, 0, height, layout)

        #  then copy to image
        self.__widget.set_from_drawable(pmap, True)
        width, height = self.__size
        if (width and height):
            self.__widget.render(width, height, a / 255.0, 1)


    def __make_label(self):

        value = self.__old_value

        self.__pango_layout.set_markup(value)
        self.__pango_layout.set_font_description(self.__font_description)
        self.__pango_layout.set_width(self.__wrap_at * pango.SCALE)
        width, height = self.__pango_layout.get_pixel_size()

        if (width and height):
            self.__render_text(self.__pango_layout)
            self.__widget.show()
        else:
            self.__widget.hide()

        self.set_size(Unit.Unit(width, Unit.UNIT_PX),
                      Unit.Unit(height, Unit.UNIT_PX))



    def __set_value(self, value):

        value = self.__utf8ify(value)
        self.__old_value = value
        self.__make_label()



    def __set_font(self, font):

        m = TargetLabel.__RE_FONT.match(font)
        if (m):
            name = m.group("name")
            unit = m.group("unit") or Unit.UNIT_PT  # pt is default for fonts
            size = m.group("size")
            if (size):
                size = float(size)
                if (unit == Unit.UNIT_PERCENT):
                    height = self._get_parent().get_geometry()[3]
                    # don't allow 0 pixels height
                    size = max(1, height.as_pt() * (size / 100.0))
                    unit = Unit.UNIT_PT
                u = Unit.Unit(size, unit)
                size = u.as_pt()
                font = "%s %f" % (name, size)

        self.__font_description = pango.FontDescription(font)
        self.__pango_context.set_font_description(self.__font_description)
        #self.__widget.set_size_request(-1, -1)


    def __set_color(self, color):

        self.__color = utils.parse_color(color)


    def __set_wrap(self, value):

        width = self._get_parent().get_geometry()[2]
        if (width > Unit.ZERO): value.set_100_percent(width.as_px())
        size = value.as_px()

        if (size == 0):
            self.__wrap_at = -1
        else:
            self.__wrap_at = size


    #
    # "value" property.
    #
    def _setp_value(self, key, value):

        if (value != self.__old_value):
            self.__set_value(value)
            self._setp(key, value)


    #
    # "font" property.
    #
    def _setp_font(self, key, value):

        if (value != self.get_prop("font")):
            self.__set_font(value)
            self._setp(key, value)
            self.__make_label()
            #self.__set_value(self.get_prop("value"))


    #
    # "color" property.
    #
    def _setp_color(self, key, value):

        self.__set_color(value)
        self._setp(key, value)
        self.__make_label()
        #self.__set_value(self.get_prop("value"))


    #
    # "wrap-at" property.
    #
    def _setp_wrap_at(self, key, value):

        self.__set_wrap(value)
        self._setp(key, value)
        self.__make_label()



from Observable import Observable
import wallpaper

import gobject
import time


#
# Class for reporting when the wallpaper image has changed.
#
class BGWatcher(Observable):

    OBS_CHANGE_BG = 0

    def __init__(self):

        if (HAVE_WIN32): return

        self.__old_bg = 0
        self.__last_update = 0.0

        gobject.timeout_add(50, self.__check_bg)



    def __check_bg(self):

        try:
            ident = wallpaper.get_wallpaper_id()
            if (ident != self.__old_bg):
                self.__last_update = time.time()
                gobject.timeout_add(500, self.__notify_update,
                                    self.__last_update)
            self.__old_bg = ident

        except StandardError:
            if (self.__old_bg):
                self.__last_update = time.time()
                gobject.timeout_add(500, self.__notify_update,
                                    self.__last_update)
                self.__old_bg = 0

        return True



    def __notify_update(self, timestamp):

        if (timestamp != self.__last_update): return True
        self.update_observer(self.OBS_CHANGE_BG)

        return False




import os

from utils import vfs

class MetaDataRegistry:
    """
    Registry for caching metadata information. By caching metadata, we can
    reduce the startup time of programs. The cache is intelligent enough to
    add uncached items and to remove obsolete ones.


    The data format of the registry file is as follows:

      - lines starting with '#' and empty lines are ignored and may be used
        for comments and formatting

      - every other line contains one item

      - an item consists of a space-separated list

      - the first entry in the list is the path of the cached item, i.e. the
        file which originally contains the metadata

      - the other entries in the list are key-value pairs, where the key and
        value are separated by a ':'

      - the value strings are URL-quoted
    """


    def __init__(self, regfile, repos, find_handler, register_handler):

        self.__regfile = regfile
        self.__repos = repos
        self.__find_handler = find_handler
        self.__register_handler = register_handler

        self.__registry = {}
        self.__loaded = {}

        try:
            self.__load()
        except:
            # don't crash when loading crap
            pass



    #
    # Scans the repositories for items, which have not yet been registered.
    #
    def scan_repositories(self):

        for repo in self.__repos:
            for path in self.__find_handler(repo):
                if (not path in self.__loaded):
                    items = self.__register_handler(path)
                    if (items):
                        self.__registry[path] = items
                        self.__loaded[path] = True
            #end for
        #end for

        # time to save
        self.__save()



    def __load(self):

        try:
            reg = open(self.__regfile).readlines()
        except:
            reg = []

        for line in reg:
            if (not line.strip() or line.startswith("#")): continue

            parts = line[:-1].split(" ")
            path = vfs.unescape_path(parts[0])
            items = parts[1:]
            self.__registry[path] = []
            for i in items:
                k, v = i.split(":")
                self.__registry[path].append((k, vfs.unescape_path(v)))
            self.__loaded[path] = True



    def __save(self):

        fd = open(self.__regfile, "w")
        fd.write(
            "# This autogenerated file acts as a cache. Removing this file "
            "will result in a\n"
            "# longer startup time.\n\n")

        for path, items in self.__registry.items():
            if (not os.path.exists(path)): continue
            data = map(lambda item: "%s:%s" % (item[0],
                                               vfs.escape_path(item[1])),
                       items)
            fd.write("%s %s\n" % (vfs.escape_path(path), " ".join(data)))

        fd.close()



    #
    # Returns a list of all items.
    #
    def get_item_list(self):

        return self.__registry.keys()



    #
    # Returns the item given by its path.
    #
    def get_item(self, path):

        return self.__registry.get(path, [])



    #
    # Sets the given item.
    #
    def set_item(self, path, metadata):

        self.__registry[path] = metadata
        self.__save()



_DELIMITER = "delimiter"
_MATCH_ONE = "?"
_MATCH_MANY = "*"


#
# Class for the node of a compressed trie.
#
class _TrieNode(object):

    __slots__ = ('__body', '__children', '__values')


    def __init__(self):
 
        # the body elements; since the trie is compressed, there may be more
        # than one
        self.__body = []

        # the children of the node
        self.__children = {}

        # the values stored in the node
        self.__values = []


    def _set_body(self, body): self.__body = body[:]
    def _set_children(self, children): self.__children = children.copy()
    def _set_values(self, values): self.__values = values[:]


    #
    # Inserts a new string into the trie.
    #
    def insert(self, data, value):
        
        # if the node is empty, just fill it
        if (not self.__body):
            self._set_body(data)
            self.__values.append(value)
            return

        index = 0
        length = len(self.__body)

        # check how far the elements match
        while (index < length):
            # split up node if elements don't match
            if (data[index] != self.__body[index]):
                ext_node = _TrieNode()
                ext_node._set_body(self.__body[index:])
                ext_node._set_children(self.__children)
                ext_node._set_values(self.__values)

                new_node = _TrieNode()
                new_node.insert(data[index:], value)
                
                self.__children = {self.__body[index]: ext_node,
                                   data[index]: new_node}
                self.__body = self.__body[:index]
                self.__values = []
                return
            #end if
            index += 1
        #end while

        # use this node if the given string exactly matches the body string
        if (index == len(data)):
            self.__values.append(value)
            return

        # insert node
        node = self.__children.get(data[index])
        if (node):
            # insert into existing node
            node.insert(data[index:], value)
        else:
            # create a new node
            new_node = _TrieNode()
            new_node.insert(data[index:], value)
            self.__children[data[index]] = new_node


    #
    # Retrieves the values which are stored for the given string. The string
    # may contain the wildcards ? (match any character) and * (match zero or
    # more characters).
    #
    # FIXME: clean up
    #
    def retrieve(self, data, case_sensitive):

        index = 0
        length = len(self.__body)

        values = []
        while (index < length):
            if (case_sensitive):
                next = data[index]
                char = self.__body[index]
            else:
                next = data[index].lower()
                char = self.__body[index].lower()
            
            if (next == _MATCH_ONE and char != _DELIMITER):
                index += 1
            elif (next == _MATCH_ONE):
                return []
            
            # the MANY wildcard can be translated to ONE wildcards
            elif (next == _MATCH_MANY):
                prefix = data[:index]
                suffix = data[index + 1:]
                values += self.retrieve(prefix + suffix, case_sensitive)
                values += self.retrieve(prefix + [_MATCH_ONE, _MATCH_MANY] +
                                        suffix, case_sensitive)
                return values
            
            elif (next != char):
                return []
            
            else:
                index += 1
        #end while

        if (index >= len(data)): return self.__values

        next = data[index]
        if (next in (_MATCH_ONE, _MATCH_MANY)):
            for node in self.__children.values():
                values += node.retrieve(data[index:], case_sensitive)
            return values
                    
        else:
            node1 = self.__children.get(next)
            if (not case_sensitive): node2 = self.__children.get(next.lower())
            else: node2 = None
            if (not case_sensitive): node3 = self.__children.get(next.upper())
            else: node3 = None

            if (node1 or node2 or node3):
                l1 = []
                l2 = []
                l3 = []
                if (node1):
                    l1 = node1.retrieve(data[index:], case_sensitive)
                if (node2):
                    l2 = node2.retrieve(data[index:], case_sensitive)
                if (node3):
                    l3 = node3.retrieve(data[index:], case_sensitive)

                return values + l1 + l2 + l3
            else:
                return values + self.__values


    #
    # Recursive method for removing entries. If it return True, then the node
    # could be removed successfully, otherwise the node is still in use.
    #
    def remove(self, key, value):

        # recursively check the children, if the key matches
        if (len(key) > len(self.__body) and
               key[:len(self.__body)] == self.__body):
            key = key[len(self.__body):]
            next = key[0]
            child = self.__children.get(next)
            if (child):
                to_remove = child.remove(key, value)
                if (to_remove):
                    del self.__children[next]
                
        # remove the value from this node
        else:
            try:
                self.__values.remove(value)
            except ValueError:
                pass

        # if the node has become empty, then we can remove it
        if (not self.__values and not self.__children):
            return True
        else:
            return False


    def get_size(self):

        size = 1
        for c in self.__children.values():
            size += c.get_size()

        return size



#
# Class for a trie.
#
class Trie:

    __slots__ = ('__root', '__case_sensitive')

    def __init__(self):

        self.__root = _TrieNode()
        self.__case_sensitive = True

        # table for storing all indices of a value: value -> [indices]
        self.__index_table = {}


    #
    # Sets the case sensitivity for searches in the trie.
    #
    def set_case_sensitive(self, value): self.__case_sensitive = value
    

    #
    # Inserts a value into the trie for the given list of key elements.
    # A list element is usually a letter of a string which is to be indexed.
    #
    def insert(self, key, value):

        key = [_DELIMITER] + key + [_DELIMITER]
        self.__root.insert(key, value)

        if (not value in self.__index_table): self.__index_table[value] = []
        self.__index_table[value].append(key)


    #
    # Retrieves the stored value for the given list of key elements. A list
    # element is usually a letter of an indexed string.
    #
    def retrieve(self, key):

        key = [_DELIMITER] + key + [_DELIMITER]
        values = self.__root.retrieve(key, self.__case_sensitive)

        if len(values) <= 1:
            return values

        # remove duplicates
        values.sort()
        new = []
        current = None
        for v in values:
            if (v != current):
                new.append(v)
                current = v
        #end for
        return new


    #
    # Removes the given value from the trie.
    #
    def remove(self, value):

        indices = self.__index_table.get(value, [])
        for index in indices:
            self.__root.remove(index, value)


    def get_size(self):

        return self.__root.get_size()




if (__name__ == "__main__"):

    text = """
    This is just a plain text which we are going to index. There are lots of
    words in it, so this will be some kind of stress test for our trie
    implementation. Let's see how well this will work with compressed tries.
    """

    import sys
    if '.' not in sys.path: sys.path.append('.')
    if '..' not in sys.path: sys.path.append('..')


    def stats():

        from libdesklets.system.gtop import proc_mem, proc_time
        from os import getpid

        pid = getpid()

        print 'CPU time: %ds\tMemory usage: %0.2f' % (proc_time(pid).utime / 100,
                                                      proc_mem(pid).resident / (1024.0**2),)
    

    stats()

    keys = text.split() #open('/usr/share/dict/words').read().split()#[:1000]

    stats()

    t = Trie()
    for word in keys:
        t.insert(list(word), word)


    stats()
    
    for key in keys:
        t.retrieve(list(key))

    stats()


    a = t.retrieve(list("w?ll*"))
    print a, len(a)

    for word in text.split():
        t.remove(word)
        print t.get_size()

    for word in keys:
        t.insert(list(word), word)
        print t.get_size()

    a = t.retrieve(list("*"))
    print a, len(a)

    stats()




from DisplayTarget import DisplayTarget
from utils.datatypes import *
from layout import Unit
from utils import vfs
from utils.tiling import Tiling
import utils

import gtk


#
# Class for image targets.
#
class TargetImage(DisplayTarget):

    def __init__(self, name, parent):

        self.__size = (0, 0)

        # the original image size
        self.__original_size = (0, 0)

        # the current saturation value
        self.__current_saturation = 0.0

        # the current opacity value
        self.__current_opacity = 0.0

        DisplayTarget.__init__(self, name, parent)

        self.__widget = Tiling()
        self.__widget.show()

        self._register_property("uri", TYPE_STRING,
                                self._setp_uri, self._getp)
        self._register_property("saturation", TYPE_FLOAT,
                                self._setp_saturation, self._getp)
        self._register_property("scale", TYPE_FLOAT,
                                self._setp_scale, self._getp)
        self._register_property("opacity", TYPE_FLOAT,
                                self._setp_opacity, self._getp)
        self._register_property("image-width", TYPE_UNIT,
                                self._setp_image_size, self._getp_image_size)
        self._register_property("image-height", TYPE_UNIT,
                                self._setp_image_size, self._getp_image_size)

        self._setp("image-width", Unit.Unit())#-1, Unit.UNIT_PX))
        self._setp("image-height", Unit.Unit())#-1, Unit.UNIT_PX))
        self._setp("opacity", 1.0)
        self._setp("saturation", 1.0)
        self._setp("scale", 1.0)

        # watch the parent for geometry changes
        self.add_observer(self.__on_observe_size)



    def get_widget(self): return self.__widget


    def delete(self):

        del self.__widget
        DisplayTarget.delete(self)



    #
    # Observer for size.
    #
    def __on_observe_size(self, src, cmd, *args):

        x, y, w, h = src.get_geometry()
        if (cmd == src.OBS_GEOMETRY and
            (w.as_px(), h.as_px()) != self.__size):
            self.__compute_size()
            utils.request_call(self.__render_image)
            self.__size = (w.as_px(), h.as_px())



    #
    # Loads the image from the given URI.
    #
    def __load_image(self, uri):

        if (not uri): return

        try:
            data = vfs.read_entire_file(uri)
        except:
            return

        try:
            self.__widget.set_from_data(data)
            self.__original_size = self.__widget.get_size()

        except ValueError, exc:
            log(`exc`)
            return



    #
    # Computes the image size.
    #
    def __compute_size(self):

        # get image-width, image-height
        imgwidth = self._getp("width")
        imgheight = self._getp("height")
        origwidth, origheight = self.__original_size
        width = imgwidth.as_px()
        height = imgheight.as_px()

        # scale aspect-ratio
        if (imgwidth.is_unset() and imgheight.is_unset()):
            width = origwidth
            height = origheight
        elif (imgwidth.is_unset() and origheight > 0):
            width = height * (origwidth / float(origheight))
        elif (imgheight.is_unset() and origwidth > 0):
            height = (width * origheight) / float(origwidth)
        
        # apply scale
        scale = self.get_prop("scale")
        # some authors don't pay attention, let's catch it
        if (scale <= 0.0): scale = 1.0
        width = int(width * scale)
        height = int(height * scale)

        # set size
        self.set_size(Unit.Unit(width, Unit.UNIT_PX),
                      Unit.Unit(height, Unit.UNIT_PX))
        utils.request_call(self.__render_image)
        

    #
    # Renders the image.
    #
    def __render_image(self):

        if (0 in self.__original_size): return

        w, h, = self.get_geometry()[2:]
        imgwidth = w.as_px()
        imgheight = h.as_px()

        if (imgwidth == 0 or imgheight == 0): return

        saturation = self.get_prop("saturation")
        opacity = self.get_prop("opacity")

        # if nothing has changed, do nothing
        #if ((imgwidth, imgheight, saturation, opacity) ==
        #    (self.__current_width, self.__current_height,
        #     self.__current_saturation, self.__current_opacity)):
        #    return

        self.__widget.render(imgwidth, imgheight, opacity, saturation)
        self.__current_width = imgwidth
        self.__current_height = imgheight
        self.__current_opacity = opacity
        self.__current_saturation = saturation



    #
    # "uri" property.
    #
    def _setp_uri(self, key, value):

        path = self._get_display().get_full_path(value)
        try:
            self.__load_image(path)

        except:
            log("Failed to set URI %s for key %s." % (path, key))
            return
        
        self.__compute_size()
        self._setp(key, value)



    #
    # "opacity" property.
    #
    def _setp_opacity(self, key, value):

        value = min(1.0, max(0.0, value))
        old_opac = self.get_prop("opacity")
        if (abs(value - old_opac) > 0.01):
            self._setp(key, value)
            utils.request_call(self.__render_image)



    #
    # "saturation" property.
    #
    def _setp_saturation(self, key, value):

        old_sat = self.get_prop("saturation")
        if (abs(value - old_sat) > 0.01 ):
            self._setp(key, value)
            utils.request_call(self.__render_image)



    #
    # "scale" property.
    #
    def _setp_scale(self, key, value):

        old_scale = self.get_prop("scale")
        if (abs(value - old_scale) > 0.01):
            self._setp(key, value)
            self.__compute_size()



    #
    # "image-width" and "image-height" properties.
    #
    def _setp_image_size(self, key, value):
        assert (isinstance(value, Unit.Unit))

        if (key == "image-width" and
              value != self._getp("image-width")):
            self.set_prop("width", value)
            self._setp(key, value)

        elif (key == "image-height" and
              value != self._getp("image-height")):
            self.set_prop("height", value)
            self._setp(key, value)

        self.__compute_size()


    #
    # "image-height" property.
    #
    def _getp_image_size(self, key):

        if (key == "image-width"):
            return self._getp(key)
        elif (key == "image-height"):
            return self._getp(key)




import os
import sys
import utils
import NullObject
import __builtin__
from utils import vfs


if (sys.platform == "win32"):
    HAVE_WIN32 = True
else:
    # no, we're lucky
    HAVE_WIN32 = False

__builtin__.HAVE_WIN32 = HAVE_WIN32

# gDesklets home paths
HOME = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
if (HAVE_WIN32):
    USERHOME = os.path.join(os.environ["APPDATA"], "gDesklets")
else:
    USERHOME = os.path.join(os.path.expanduser("~"), ".gdesklets")
OLDHOME = os.path.join(HOME, os.pardir, os.pardir, "share", "gdesklets")


# we need the DISPLAY variable
if (HAVE_WIN32):
    DISPLAY = "win32"
else:
    try:
        DISPLAY = vfs.escape_path(os.environ["DISPLAY"].replace("/", "_"))
    except KeyError:
        print "The DISPLAY variable is NOT set, which usually means, " \
              "that X isn't running!"
        sys.exit(1)


# the name of the purge key
PURGE_KEY = "_paths_to_purge_"

# for unset coordinates
UNSET_COORD = -1000

# paths where the display files are located
DISPLAYPATHS = (os.path.join(USERHOME, "Displays"),
                os.path.join(HOME, "Displays"),
                os.path.join(OLDHOME, "Displays"))
# paths where gDesklets looks for sensors and displays
SENSORPATHS = (os.path.join(USERHOME, "Sensors"),
               os.path.join(HOME, "Sensors"),
               os.path.join(OLDHOME, "Sensors"))
# paths where gDesklets looks for controls
CONTROLPATHS = (os.path.join(USERHOME, "Controls"),
                os.path.join(HOME, "Controls"),
                os.path.join(OLDHOME, "Controls"))
# path to the log file(s)
LOG_PATH = os.path.join(USERHOME, "logs")
# log file
LOGFILE = os.path.join(LOG_PATH, "gdesklets" + DISPLAY + ".log")
# path for registry files
REGISTRY_PATH = os.path.join(USERHOME, "registry")
# path of the communication sockets
SOCKET_PATH = os.path.join(USERHOME, "sockets")
# path containing gDesklets' PID file
PID_PATH = os.path.join(USERHOME, "gdesklets%s.pid" % DISPLAY)

# setup i18n
from utils.i18n import Translator
_ = Translator("gdesklets")


NAME = "gDesklets"
VERSION = "0.35.1"
COPYRIGHT = u"Copyright \xa9 2003 - 2005 The gDesklets Team"
DESCRIPTION = _("A desktop applet system for GNOME")
AUTHORS = ( "Martin Grimme <martin@pycage.de>",
            "Christian Meyer <chrisime@gnome-de.org>",
            "Jesse Andrews <jdandr2@cs.uky.edu>",
           u"S\xe9bastien Bacher <seb128@debian.org>",
           u"Beno\xeet Dejean <tazforever@dlfp.org>"
          )
DOCUMENTERS = ("Martin Grimme <martin@pycage.de>",)
ICON = os.path.join(HOME, "data", "gdesklets.png")
CREDITS = (
           (_("Version %s") % (VERSION,),
           (u"Copyright \xa9 2003 - 2005", "The gDesklets Team")),

           (_("Core Programming:"),
           ("Martin Grimme", "Christian Meyer", "Jesse Andrews",
            u"Beno\xeet Dejean")),

           (_("Architecture Design:"),
           ("Martin Grimme", "Christian Meyer")),

           (_("GNOME Integration:"),
           ("Christian Neumair", "Martin Grimme", u"S\xe9bastien Bacher",
            "Christian Kellner", "Christian Meyer")),

           (_("Artwork:"),
           ("Johannes Rebhan",)),

           (_("Website:"),
           ("Luke Stroven",)),

           (_("Thank you:"),
           ("VidaLinux for sponsoring",
            "the http://www.gdesklets.org domain!",))
          )


def init():

    """ gDesklets init process """
    import signal
    import gtk

    # prepare GTK for threads
    try:
        gtk.threads_init()
    except:
        print "No threading support available in python."
        print "Compile python with --enable-threads. Exiting!"
        sys.exit(1)

    # install signal handler to quit on Ctrl-C
    signal.signal(signal.SIGINT, gtk.main_quit)




from libdesklets.controls import Control

from ITime import ITime

import time
import os


#
# Control for time-related stuff.
#
class Time(Control, ITime):

    def __init__(self):

        self.__timezone = ""
        self.__timezone_offset = time.timezone

        Control.__init__(self)

        self.__yesterday = None

        self._add_timer(1000, self.__tick_time)
        

    def __tick_time(self):

        self._update("time")
        self.__try_tick_date()
        return True


    def __try_tick_date(self):

        today = self.__get_date()
        
        if today != self.__yesterday:
            # all my troubles seem so far away
            self.__yesterday = today
            self._update("date")



    def __set_timezone(self, tz):

        self.__timezone = tz
        #have_tz = "TZ" in os.environ

        #old_tz = os.environ.get("TZ", "")
        #if (self.__timezone):
        #    os.environ["TZ"] = self.__timezone
        #    time.tzset()
            
        #self.__timezone_offset = time.timezone

        #if (self.__timezone):
        #    if (not have_tz): del os.environ["TZ"]
        #    else: os.environ["TZ"] = old_tz
        #    time.tzset()

        #self.__try_tick_date()
        self._update("timezone")
        #self._update("date")


    def __get_timezone(self): return self.__timezone


    def __get_time_and_date(self):

        if (self.__timezone):
            have_tz = "TZ" in os.environ
            old_tz = os.environ.get("TZ", "")

            os.environ["TZ"] = self.__timezone
            time.tzset()

            tme = time.localtime()

            if (not have_tz): del os.environ["TZ"]
            else: os.environ["TZ"] = old_tz
            time.tzset()
        else:
            tme = time.localtime()
            
        #tme = time.gmtime(time.time() - self.__timezone_offset)            
        return tme


    def __get_time(self):

        year, month, day, hours, minutes, seconds, weekday, julian, dsflag = \
               self.__get_time_and_date()
        return (hours, minutes, seconds)


    def __get_date(self):

        year, month, day, hours, minutes, seconds, weekday, julian, dsflag = \
               self.__get_time_and_date()
        return (year, month, day)



    timezone = property(__get_timezone, __set_timezone, doc = "the timezone")
    time     = property(__get_time, doc = "the current time (h, m, s)")
    date     = property(__get_date, doc = "the current date (y, m, d)")



def get_class(): return Time



from ContainerTarget import ContainerTarget
from utils.datatypes import *
from utils.Struct import Struct
from layout import Unit
from utils import vfs

import gtk
import os


_STATE_VISIBLE = 0
_STATE_HIDDEN = 1


#
# Class for target expander.
#
class TargetExpander(ContainerTarget):

    __ACTION_EXPAND = "expand"

    def __init__(self, name, parent):

        self.__show_state = _STATE_VISIBLE

        ContainerTarget.__init__(self, name, parent)
        self.__widget = gtk.Table(2, 2)
        self.__widget.show()

        self.__icon = gtk.Image()
        self.__icon.show()
        self.__widget.attach(self.__icon, 0, 1, 0, 1, 0, 0, 0, 0)

        dummy = gtk.HBox(); dummy.show()
        self.__widget.attach(dummy, 0, 1, 1, 2,
                             gtk.EXPAND | gtk.FILL, gtk.EXPAND | gtk.FILL,
                             0, 0)

        self._register_property("expanded", TYPE_BOOL,
                                 self._setp_expanded, self._getp)
        self._register_property("icon-uri", TYPE_STRING,
                                 self._setp_icon, self._getp)
        self._register_action(self.__ACTION_EXPAND)


    def get_widget(self): return self.__widget


    def __on_click(self, src, event):

        if (self.__show_state == _STATE_VISIBLE):
            self.__set_expanded(False)
        else:
            self.__set_expanded(True)


    def __set_expanded(self, value):

        if (value): self.__show_state = _STATE_VISIBLE
        else: self.__show_state = _STATE_HIDDEN

        iconw, iconh = self.__icon.size_request()

        if (self.__show_state == _STATE_VISIBLE):
            self._get_child().set_prop("visible", True)
            width, height = self._get_child().get_geometry()[2:]
            self.set_size(iconw + width.as_px(), height.as_px())
            expand = True
        else:
            self._get_child().set_prop("visible", False)
            self.set_size(iconw, iconh)
            expand = False

        self._get_display().send_action(self, self.__ACTION_EXPAND,
                                        Struct(expand = expand,
                                               _args = [expand]))


    def __set_icon(self, uri):

        if (not uri): return
        uri = self._get_display().get_full_path(uri)
        loader = gtk.gdk.PixbufLoader()
        try:
            data = vfs.read_entire_file(uri)
        except:
            return
        try:
            loader.write(data, len(data))
        except:
            log("Invalid image format.")
            return

        loader.close()
        pbuf = loader.get_pixbuf()
        self.__icon.set_from_pixbuf(pbuf)



    def get_border_size(self):

        w, h = self.__icon.size_request()
        if (self.__show_state == _STATE_VISIBLE):
            return (Unit.Unit(w, Unit.UNIT_PX), Unit.ZERO,
                    Unit.ZERO, Unit.ZERO)
        else:
            return (Unit.Unit(w, Unit.UNIT_PX), Unit.Unit(h, Unit.UNIT_PX),
                    Unit.ZERO, Unit.ZERO)


    def _setp_expanded(self, key, value):

        self.__set_expanded(value)
        self._setp(key, value)


    def _setp_icon(self, key, value):

        self.__set_icon(value)
        self._setp(key, value)


    def get_container_geometry(self):

        if (self._get_child()):
            x, y, w, h = self._get_child().get_geometry()
            iconw, iconh = self.__icon.size_request()
        else:
            x, y, w, h = self.get_geometry()

        return (x, y, w, h)


    def handle_action(self, action, px, py, args):

        ContainerTarget.handle_action(self, action, px, py, args)

        w, h = self.__icon.size_request()
        if (0 <= px.as_px() <= w and 0 <= py.as_px() <= h):
            if (action == self.ACTION_CLICK):
                if (self.__show_state == _STATE_VISIBLE):
                    self.__set_expanded(False)
                else:
                    self.__set_expanded(True)


    def new_child(self, childtype, settings, children):

        child = ContainerTarget.new_child(self, childtype, settings, children)
        self.__widget.attach(child.get_widget(), 1, 2, 0, 2,
                             gtk.EXPAND | gtk.FILL, gtk.EXPAND | gtk.FILL,
                             0, 0)



from shell.Plugin import Plugin
from main import USERHOME, HOME
from main.DisplayList import DisplayList
from main import client

import os
import sys


#
# Plugin for acting as a client for the gDesklets daemon.
#
class Core_Client(Plugin):

    def init(self):
        self.__daemon = None
        self.__dsplist = DisplayList(os.path.join(USERHOME, "displays"))
        self.__get_daemon()
        profile = self.__dsplist.get_profile()
        self.set_profile(profile)


    def __get_daemon(self):

        self.__daemon = client.get_daemon()
        self.__daemon.set_remove_command(os.path.join(HOME, "gdesklets") +
                                         " _remove")


    def open_display(self, path):

        try:
            ident =  self.__daemon.open_display(path)
        except:
            self.__get_daemon()
            ident =  self.__daemon.open_display(path)

        profile = self.__dsplist.get_profile()
        self.__dsplist.add_display(profile, path, ident)
        self.__dsplist.commit()

        return ident


    def open_display_with_id(self, path, ident):

        try:
            ret = self.__daemon.open_display_with_id(path, ident)
        except:
            self.__get_daemon()
            ret = self.__daemon.open_display_with_id(path, ident)
        return ret


    def close_display(self, ident):

        try:
            self.__daemon.close_display(ident)
        except:
            self.__get_daemon()
            self.__daemon.close_display(ident)


    def get_profile(self):

        return self.__dsplist.get_profile()


    def get_profiles(self):

        return self.__dsplist.get_profiles()


    def set_profile(self, profile):

        # close displays
        current_profile = self.__dsplist.get_profile()
        displays = self.__dsplist.get_displays(current_profile)
        if (current_profile != profile):
            for ident in displays:
                try:
                    self.__daemon.close_display(ident)
                except:
                    self.__get_daemon()
                    self.__daemon.close_display(ident)
        #end if

        self.__dsplist.set_profile(profile)
        self.__dsplist.commit()

        # open displays
        displays = self.__dsplist.get_displays(profile)
        for ident in displays:
            try:
                path = self.__dsplist.lookup_display(ident)[-1]
            except KeyError, exc:
                print >> sys.stderr, exc
            try:
                self.__daemon.open_display_with_id(path, ident)
            except:
                self.__get_daemon()
                self.__daemon.open_display_with_id(path, ident)


    def get_displays(self, profile):

        displays = []
        for ident in self.__dsplist.get_displays(profile):
            try:
                profile, path = self.__dsplist.lookup_display(ident)
            except KeyError, exc:
                print >> sys.stderr, exc

            displays.append((ident, path))

        return displays


def get_class(): return Core_Client



import gobject
import gtk


class Downloader(gtk.Dialog):

    def __init__(self, path):

        self.__is_cancelled = False

        gtk.Dialog.__init__(self, title = "", buttons = (gtk.STOCK_CANCEL,
                                                         gtk.RESPONSE_CANCEL))
        self.set_default_size(300, 100)
        self.connect("response", self.__on_cancel)

        vbox = self.vbox

        hbox = gtk.HBox(False, 12)
        hbox.set_border_width(12)
        vbox.pack_start(hbox, True, True)
        import os
        img = gtk.Image()
        img.set_from_file(os.path.join(path, "download.png"))
        hbox.pack_start(img, False, False)

        vbox = gtk.VBox()
        hbox.pack_end(vbox, True, True)

        lbl = gtk.Label("<b>" + _("Retrieving:") + "</b>")
        lbl.set_use_markup(True)
        align = gtk.Alignment(0.0, 0.0)
        align.add(lbl)
        vbox.add(align)

        self.__label = gtk.Label("")
        self.__label.set_use_markup(True)
        align = gtk.Alignment(0.0, 0.0)
        align.add(self.__label)
        vbox.add(align)

        self.__bar = gtk.ProgressBar()
        vbox.add(self.__bar)


    def __on_cancel(self, src, response):

        self.__is_cancelled = True


    def download(self, url, dest):

        name = url
        if (len(name) >= 60):
            name = name[:30] + "..." + name[-30:]
        gobject.timeout_add(0, self.__label.set_text, "%s" % (name))
        gobject.timeout_add(0, self.__bar.set_fraction, 0)
        gobject.timeout_add(0, self.__bar.set_text, "Contacting...")

        gobject.timeout_add(0, self.show_all)
        self.__is_cancelled = False

        dest_fd = open(dest, "w")
        import urllib2
        src_fd = urllib2.urlopen(url)
        total_size = src_fd.info().get("Content-Length", 0)
        so_far = 0
        while (not self.__is_cancelled):
            data = src_fd.read(4096)
            if (not data):
                break
            dest_fd.write(data)
            so_far += len(data)
            value = (100 * so_far / max(0.1, float(total_size)))
            gobject.timeout_add(0, self.__bar.set_fraction, value / 100.0)
            gobject.timeout_add(0, self.__bar.set_text, "%i%%" % (value))

        src_fd.close()
        dest_fd.close()

        gobject.timeout_add(0, self.hide)




from sensor.Sensor import Sensor
from utils.datatypes import *
from utils import i18n
import time
import os


#
# FontSelector sensor by Psi (psiz _at_ free _dot_ fr)
# Adopted and modified by Martin Grimme.
#
class FontSelector(Sensor):

    def __init__(self, fontnumber="1", *fontargs):

        global _; _ = i18n.Translator("font-selector-sensor")


        # the number of fonts
        self.__fontnumber = int(fontnumber)

        # the font settings
        self.__fonts = []

        # the color settings
        self.__colors = []


        Sensor.__init__(self)
        self._watch_config(self.config_watcher)
        # setup configuration entries
        if (len(fontargs) == 1 and fontargs[0].find("  ") != -1):
            fontargs = \
               (fontargs[0].replace("   ", ",").replace("  ", ",")).split(",")
            print "Using spaces to separate font entries is deprecated."
            print "Use commas instead."

        fontargs = list(fontargs)
        for font_n in xrange(self.__fontnumber):
            if (fontargs): font = fontargs.pop(0)
            else: font = "Sans 8"
            if (fontargs): color = fontargs.pop(0)
            else: color = "black"
            self._set_config_type("font%i" % font_n, TYPE_STRING, font)
            self._set_config_type("color%i" % font_n, TYPE_STRING, color)
        #end for

        self._add_timer(500,self.__font_init)

    def config_watcher(self,key,value):
        data = self._new_output()
        data.set(key, value)
        self._send_output(data)

    #
    # Font init
    #
    def __font_init(self):

        fonts = []
        colors = []

        for font_n in xrange(self.__fontnumber):
            fonts.append(self._get_config("font%i" % font_n))
            colors.append(self._get_config("color%i" % font_n))


        if (fonts != self.__fonts or colors != self.__colors):
            self.__fonts = fonts
            self.__colors = colors

            data = self._new_output()
            for font_n in xrange(self.__fontnumber):
                data.set("font%i" % font_n, fonts[font_n])
                data.set("color%i" % font_n, colors[font_n])
            self._send_output(data)
        #end if

        return 0



    def get_configurator(self):

        configurator = self._new_configurator()
        configurator.set_name(_("Fonts"))
        configurator.add_title(_("Fonts settings"))

        for font_n in xrange(self.__fontnumber):
            configurator.add_font_selector(_("Font #%i:" % font_n),
                                           "font%i" % font_n,
                               _("Font #%i used in the labels" % font_n))
            configurator.add_color_selector(_("Font color #%i:" % font_n),
                                            "color%i" % font_n,
                               _("Font color #%i used in the labels" % font_n))
        #end for

        return configurator



def new_sensor(args): return FontSelector(*args)



#
# Class for encapsulating scriptlets. A scriptlet is a part of an inline script.
# Each <script> tag defines one scriptlet. So do the action handlers. The main
# purpose of this class is to store metadata together with scriptlets, in order
# to be able to identify them.
#
class Scriptlet:

    def __init__(self, script, filename):

        self.__script = script
        self.__filename = filename

    def __get_script(self): return self.__script
    def __get_filename(self): return self.__filename

    def __get_id(self):

        ident = "%s_%d" % (self.__filename, hash(self.__script))
        return ident



    script = property(__get_script)
    filename = property(__get_filename)
    
    # each scriptlet has an ID; two different but identical scriptlet instances
    # share the same ID, thus making scriptlets comparable
    script_id = property(__get_id)



import gtk
import pango
import textwrap


# attributes for rendering text
_ATTR_TITLE =      {"weight": pango.WEIGHT_BOLD,
                    "font": "Mono"}
_ATTR_DEBUG =      {"foreground": "white",
                    "background": "red",
                    "weight": pango.WEIGHT_BOLD,
                    "underline": pango.UNDERLINE_SINGLE,
                    "font": "Mono"}
_ATTR_DEPRECATED = {"foreground": "black",
                    "background": "yellow",
                    "weight": pango.WEIGHT_BOLD,
                    "font": "Mono"}
_ATTR_ERROR =      {"foreground": "darkred",
                    "background": "white",
                    "weight": pango.WEIGHT_BOLD,
                    "font": "Mono"}
_ATTR_INFO =       {"foreground": "black",
                    "background": "green",
                    "weight": pango.WEIGHT_BOLD,
                    "font": "Mono"}
_ATTR_WARN =       {"foreground": "black",
                    "background": "orange",
                    "weight": pango.WEIGHT_BOLD,
                    "font": "Mono"}
_ATTR_CODE =       {"foreground": "navy",
                    "background": "white",
                    "font": "Mono"}
_ATTR_BAD_CODE =   {"foreground": "navy",
                    "background": "#ff7173",
                    "font": "Mono"}
_ATTR_NORMAL =     {"foreground": "black",
                    "background": "white"}


#
# A specialized TextView widget for formatting log and error messages with
# markups.
#
class LogView(gtk.TextView):

    def __init__(self):

        self.__wrap_at = 85  # 85 to have line numbers before code lines

        gtk.TextView.__init__(self)
        self.set_editable(False)
        self.set_wrap_mode(gtk.WRAP_NONE)



    def append(self, data):

        """ Adds data to the buffer """

        lines = data.splitlines()
        for line in lines:
            text, attrs = self.__markup(line)
            for chunk in textwrap.wrap(text, self.__wrap_at):
                chunk = chunk.ljust(self.__wrap_at) + "\n"

                end_iter = self.get_buffer().get_end_iter()
                tag = self.get_buffer().create_tag(None, **attrs)
                self.get_buffer().insert_with_tags(end_iter, chunk, tag)



    def __markup(self, text):

        """ Analyzes the given text and returns the appropriate attribute. """

        if ("Deprecation" in text):
            return (text, _ATTR_DEPRECATED)
        elif (text.startswith("===")):
            return (text, _ATTR_TITLE)
        elif (text.startswith("[EXC]")):
            return (text[5:], _ATTR_ERROR)
        elif (text.startswith("[DSC]")):
            return (text[5:], _ATTR_INFO)
        elif ("Info:" in text):
            return (text, _ATTR_INFO)
        elif ("Debug:" in text):
            return (text, _ATTR_DEBUG)
        elif ("Warning:" in text):
            return (text, _ATTR_WARN)
        elif (text.startswith("[---]")):
            return (text[5:], _ATTR_CODE)
        elif (text.startswith("[ERR]")):
            return (text[5:], _ATTR_BAD_CODE)
        else:
            return (text, _ATTR_NORMAL)



from ConfigWidget import ConfigWidget
from utils.datatypes import *

import gtk


class ConfigButton(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        self.__initialized = False

        ConfigWidget.__init__(self, name, getter, setter, caller)


    def update(self): pass


    def get_widgets(self):

        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        self.__button = gtk.Button("")
        self.__button.show()
        self.__button.connect("clicked", self.__on_click)
        align.add(self.__button)

        return (align,)


    def __on_click(self, src):

        self.call_callback()


    def _set_label(self, value): self.__button.set_label(value)
    def _set_enabled(self, value): self.__button.set_sensitive(value)



from ConfigWidget import ConfigWidget
from utils.datatypes import *

import gtk


class ConfigEnum(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        self.__items_values = []

        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("items", TYPE_LIST, self._setp_items,
                                self._getp, [],
                                doc = "List of (label, value) items")

        self._register_property("selection", TYPE_INT, self._setp_selection,
                                self._getp, 0,
                                doc = "Current selection index")

        self._register_property("value", TYPE_STRING, self._setp_value,
                                self._getp, "",
                                doc = "Current selection")



    def get_widgets(self):

        self.__label = gtk.Label("")
        self.__label.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(self.__label)

        value = self._get_config()
        self.__optmenu = gtk.combo_box_new_text()
        self.__optmenu.show()
        self.__optmenu.connect("changed", self.__on_change)
        
        return (align, self.__optmenu)



    def _set_enabled(self, value):

        self.__optmenu.set_sensitive(value)


    def _set_label(self, value):

        self.__label.set_text(value)
        

    def __on_change(self, src):

        if (self.__items_values):
            value = self.__items_values[src.get_active()]
            self._setp("selection", self.__optmenu.get_active())
            self._set_config(value)


    def _setp_items(self, key, items):

        self.__items_values = []
        self.__optmenu.get_model().clear()
        for k, v in items:
            self.__optmenu.append_text(k)
            self.__items_values.append(v)
        #end for

        self._setp(key, items)
        self.set_prop("value", self.get_prop("value"))


    def _setp_selection(self, key, value):

        self.__optmenu.set_active(value)
            

    def _setp_value(self, key, value):

        try:
            index = self.__items_values.index(value)
        except:
            index = 0

        self.__optmenu.set_active(-1)  # make sure to get a change event
        self.__optmenu.set_active(index)
        self._setp(key, value)



import os
import gettext
import sys


def find_locale_dir():

    fullpath = os.path.abspath(sys.argv[0])
    d, f = os.path.split(fullpath)

    if f == 'gdesklets':
        # we're in $prefix/bin/
        return os.path.join(d, os.pardir, 'share', 'locale')

    elif f in ('gdesklets-daemon', 'gdesklets-shell', 'gdesklets-logview'):
        # we're in $prefix/lib/gdesklets
        return os.path.join(d, os.pardir, os.pardir, 'share', 'locale')


def Translator(domain):
    try:
        localedir = find_locale_dir()
        return gettext.translation(domain, localedir).gettext
    except IOError:
        return lambda s: s



#
# Parser for action strings.
#
# Grammar:
# ========
# S     -> CALL | CALL , S
# CALL  -> id : call ARGS | call ARGS
# ARGS  -> (ARGS') | () | e
# ARGS' -> arg | arg , ARGS'
#


def parse(rest):

    cmds = []
    while (rest):
        cmd, rest = _parse_command(rest)
        cmds.append(cmd)
    #end while

    return cmds



def _parse_command(cmd):

    inf = len(cmd) + 1
    index1 = cmd.find("(")
    if (index1 == -1): index1 = inf
    index2 = cmd.find(",")
    if (index2 == -1): index2 = inf

    index = min(index1, index2)
    parts, rest = cmd[:index], cmd[index:]
    if (":" in parts):
        ident, call = parts.split(":")
    else:
        ident, call = None, parts

    if (rest and rest[0] == "("):
        args, rest = _parse_args(rest)
    else:
        args = []

    index = rest.find(",")
    if (index != -1): rest = rest[index + 1:]

    if (ident): ident = ident.strip()
    call = call.strip()
    return ((ident, call, args), rest)



def _parse_args(rest):

    args = []
    while (rest and rest[0] != ")"):
        arg, rest = _parse_arg(rest[1:])
        if (arg): args.append(arg)
    #end while

    return (args, rest[1:])



def _parse_arg(cmd):

    # states are: 0 none
    #             1 read " string
    #             2 escape "
    #             3 read ' string
    #             4 escape '
    #             5 escape

    arg = ""
    state = 0
    for pos in xrange(len(cmd)):
        c = cmd[pos]
        if (state == 0):
            if (c == "\""): arg += c; state = 1
            elif (c == "'"): arg += c; state = 3
            elif (c == "\\"): arg += c; state = 5
            elif (c == ","): break
            elif (c == ")"): break
            else: arg += c
        elif (state == 1):
            if (c == "\""): arg += c; state = 0
            elif (c == "\\"): arg += c; state = 2
            else: arg += c
        elif (state == 2): arg += c; state = 1
        elif (state == 3):
            if (c == "'"): arg += c; state = 0
            elif (c == "\\"): arg += c; state = 4
            else: arg += c
        elif (state == 4): arg += c; state = 3
        elif (state == 5): arg += c; state = 0
        #end if
    #end for
    if (state != 0): print "Error"

    rest = cmd[pos:]
    arg = arg.strip().replace("\\,", ",").replace("\\)", ")")
    return (arg, rest)



from shell.Plugin import Plugin
from main import ICON

import gtk

from utils import vfs


#
# The GUI shell where other plugins can plug into.
#
class UI_Shell(Plugin,):

    __TARGET_URL = 0
    __NETSCAPE_URL = 1

    # what we accept for drag & drop
    __DND_TYPES = [("text/uri-list", 0, __TARGET_URL),
                   ("x-url/http", 0, __NETSCAPE_URL),
                   ("_NETSCAPE_URL", 0, __NETSCAPE_URL)]


    def init(self):
        win = gtk.Window()
        win.set_title("gDesklets Shell")
        win.set_default_size(500, 300)
        win.connect("delete-event", self.__on_close)

        # set the icon
        win.set_icon(gtk.gdk.pixbuf_new_from_file(ICON))

        vbox = gtk.VBox()
        vbox.set_border_width(2)
        win.add(vbox)

        self.__menubox = gtk.HBox()
        vbox.pack_start(self.__menubox, False, False, 0)

        self.__toolbarbox = gtk.HandleBox()

        pane = gtk.HPaned()
        vbox.add(pane)

        lpane = gtk.VBox()
        pane.add1(lpane)

        self.__sidebarbox = gtk.HBox()
        lpane.pack_start(self.__sidebarbox, True, True, 0)

        self.__switchbox = gtk.HBox()
        lpane.pack_end(self.__switchbox, False, False, 0)

        self.__bodybox = gtk.HBox()
        pane.add2(self.__bodybox)

        self.__statusbox = gtk.HBox()
        vbox.pack_end(self.__statusbox, False, False, 0)

        win.show_all()

        menu = self._get_plugin("UI_Menu")
        menu.set_item("File", None, _("_File"), None)
        menu.set_slot("File/Slot")
        menu.set_separator("File/Separator")
        menu.set_item("File/Quit", gtk.STOCK_QUIT, "", self.__quit)

        menu.set_slot("Slot1")
        menu.set_slot("Slot2")

        plugins = self._get_plugins_by_pattern("name", "Shell_*")
        installer = self._get_plugin("Installer_Package")

        # set up drag & drop
        self.__bodybox.drag_dest_set(gtk.DEST_DEFAULT_ALL, self.__DND_TYPES,
                                     gtk.gdk.ACTION_COPY)
        self.__bodybox.connect("drag_data_received", self.__on_file_drop)



    def __on_file_drop(self, widget, context, x, y, data, info, time):

        files = [uri for uri in data.data.split("\r\n") if uri != '']
        installer = self._get_plugin("Installer_Package")

        if (info == self.__NETSCAPE_URL):
            for f in files:
                f, name = f.splitlines()[0:2]
                installer.install_from(f)
            #end for

        elif (info == self.__TARGET_URL):
            for f in files:
                if (f.startswith("file://")): f = f[7:]
                # filenames are URL-quoted
                installer.install_from(vfs.unescape_path(f))
            #end for



    def __on_close(self, src, event):

        self.__quit()


    def set_menu(self, widget):

        self.__menubox.add(widget)


    def set_toolbar(self, widget):

        self.__toolbarbox.add(widget)


    def set_switch(self, widget):

        self.__switchbox.add(widget)


    def set_sidebar(self, widget):

        self.__sidebarbox.add(widget)


    def set_body(self, widget):

        self.__bodybox.add(widget)


    def set_statusbar(self, widget):

        self.__statusbox.add(widget)


    def run(self):

        gtk.main()


    def __quit(self):

        gtk.main_quit()


def get_class(): return UI_Shell


try:
    import pygtk; pygtk.require("2.0")
except:
    pass

gtk.threads_init()




import ArchFactory

if (not HAVE_WIN32):
    from gtop import *

    # deprecated
    proc_list = proclist
    net_load = netload


_arch = ArchFactory.create_arch()

net_devices  = _arch.net_devices
net_speed    = _arch.net_speed
net_state    = _arch.net_state

cpu_bogomips = _arch.cpu_bogomips
cpu_cache    = _arch.cpu_cache
cpu_load     = _arch.cpu_load
cpu_model    = _arch.cpu_model
cpu_speed    = _arch.cpu_speed

kernel_version    = _arch.kernel_version
hostname          = _arch.hostname
operating_system  = _arch.operating_system
users             = _arch.users



#----------------------------------------------------------------------------------#
#
# FILE:                 __init__.py
# AUTHOR:               Peter Quinn
# DATE:                 20.02.2005
# VERSION:              1.30
#
# DESCRIPTION:  A object buffer that makes it easier to scroll and update arrays of
#                               objects.
#
# INTERFACE:   read           R     Read the window view array
#              read_all       R     Read the entire buffer array
#              delete         W     Deletes one line from the buffer
#              fill           W     Fills the buffer with an object
#              write          W     Write an array to the buffer
#              filltype       RW    Get / Set buffer fill type with an object
#              window_pos     RW    Get / Set viewable window position
#              window_size    RW    Get / Set viewable window size
#              size           RW    Get / Set buffer size
#              cursor         RW    Get / Set cursor position
#
#----------------------------------------------------------------------------------#

from libdesklets.controls import Interface, Permission

#----------------------------------------------------------------------------------#
#
# IArrayBuffer Interface
#
class IArrayBuffer(Interface):

    read        = Permission.READ
    read_all    = Permission.READ
    delete      = Permission.WRITE
    fill        = Permission.WRITE
    write       = Permission.WRITE
    filltype    = Permission.READWRITE
    window_pos  = Permission.READWRITE
    window_size = Permission.READWRITE
    size        = Permission.READWRITE
    cursor      = Permission.READWRITE




from libdesklets.controls import Control

import os

from utils import vfs

from IURI import IURI

class URI(Control, IURI):

    def __init__(self):

        self.__data = ""
        self.__fd   = None
        self.__list = []
        self.__uri  = ""

        Control.__init__(self)



    def __get_file(self):

        return self.__uri

    def __set_file(self, val):

        self.__uri = val
        try:
            self.__fd = vfs.open(self.__uri, OPEN_READ)
        except:
            raise IOError("Could not open file %s." % (self.__uri))

        self._update("uri")



    def __read_file(self):

        while (True):
            try:
                d = self.__fd.read(1024)
            except:
                break
            if (not d): break
            else: self.__data += d



    def __get_raw_content(self):

        self.__read_file()
        return self.__data



    def __get_strip_content(self):

        self.__read_file()
        self.__list.extend( self.__data.splitlines() )
        return self.__list



    def __get_splitted_content(self):

        self.__read_file()
        lines = self.__data.strip()
        self.__list = lines.split()
        return self.__list


    file  = property(__get_file, __set_file, doc="raw file content")
    raw   = property(__get_raw_content, doc="raw file content")
    stripped = property(__get_strip_content, doc="stripped file content")
    splitted = property(__get_splitted_content, doc="splitted file content")


def get_class(): return URI



from shell.Plugin import Plugin

import gtk


#
# Plugin for switching profiles.
#
class Shell_Profiles(Plugin):

    def init(self):
        self.__is_blocked = False
        self.__items = []
        self.__client = self._get_plugin("Core_Client")

        self.__menu = self._get_plugin("UI_Menu")
        self.__menu.insert("Slot1", "Profiles")
        self.__menu.set_item("Profiles", None, _("_Profiles"), None)
        self.__menu.set_item("Profiles/New", gtk.STOCK_NEW,
                             _("New profile..."), self.__new_profile)
        self.__menu.set_separator("Profiles/Separator")
        self.__menu.set_slot("Profiles/Slot")

        self.__build_menu()


    def __build_menu(self):

        for item, profile in self.__items:
            self.__menu.remove_item(item)

        self.__items = []
        profiles = self.__client.get_profiles()
        profiles.sort()
        profiles.reverse()
        for p in profiles:
            item = "Profiles/" + p
            self.__menu.insert("Profiles/Slot", p)
            self.__menu.set_check_item(item, p, self.__switch, p)
            self.__items.append((item, p))

        self.__set_checks()


    def __set_checks(self):

        self.__is_blocked = True

        profile = self.__client.get_profile()

        for item, p in self.__items:
            self.__menu.set_checked(item, (p == profile))

        self.__is_blocked = False


    def __switch(self, profile):

        current_profile = self.__client.get_profile()

        if (self.__is_blocked or profile == current_profile):
            self.__set_checks()
            return

        self.__client.set_profile(profile)
        self.__set_checks()


    def __new_profile(self):

        from utils.HIGDialog import HIGDialog

        dialog = HIGDialog((gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                            gtk.STOCK_OK, gtk.RESPONSE_OK),
                           gtk.STOCK_NEW, _("New profile"),
                           _("Enter the name for a new profile"))

        align = gtk.Alignment(0.0, 0.0, 1.0, 0.0)
        align.set_property("border-width", 6)
        entry = gtk.Entry()
        align.add(entry)

        def response(dialog, response):
            if (response == gtk.RESPONSE_OK):
                self.__create_profile(entry.get_text())

            dialog.destroy()

        dialog.connect("response", response)
        dialog.vbox.pack_end(align, False, False, 0)
        dialog.show_all()


    def __create_profile(self, profile):

        self.__switch(profile)
        self.__build_menu()


def get_class(): return Shell_Profiles



from utils.Element import Element
from utils.datatypes import *

import gobject


class ConfigWidget(Element):
    """
      Abstract base class for configuration widgets. Configuration widgets can
      be directly addressed to get/set values or can be visualized to allow the
      user to change settings.
    """

    def __init__(self, name, getter, setter, caller):

        Element.__init__(self, name)

        self.__may_set_back = False
        self.__initialized = False

        self.__path = ""

        self.__getter = getter
        self.__setter = setter
        self.__caller = caller
        self.__bind = ""

        self._register_property("bind", TYPE_STRING, self._setp_bind, None,
                                "",
                                doc = "Binds the value to an object property")
        self._register_property("callback", TYPE_STRING,
                                self._setp, self._getp, None,
                                doc = "Callback function")
        self._register_property("help", TYPE_STRING, self._setp, self._getp,
                                "", doc = "Tooltip text")
        self._register_property("label", TYPE_STRING,
                                self._setp_label, self._getp, "",
                                doc = "Description label text")
        self._register_property("enabled", TYPE_BOOL,
                                self._setp_enabled, self._getp, True,
                                doc = "Whether the widget is enabled")


    def _get_path(self): return self.__path
    def set_path(self, path): self.__path = path
    


    #
    # Tells the config widget to update itself by reading from the bound object.
    #
    def update(self):

        self.__may_set_back = True

        old_value = self.get_prop("value")
        try:
            new_value = self._get_config()
        except KeyError:
            return

        if (not self.__initialized or old_value != new_value):
            self.__initialized = True
            self.set_prop("value", new_value)



    def _get_config(self):

        return self.__getter(self.__bind)


    def _set_config(self, v):
        assert self.__bind

        if (not self.__may_set_back): return

        # get_datatype_of_property() doesn't throw here, since ConfigWidget has
        # a "value" property
        datatype = self.get_datatype_of_property("value")
        self.__setter(self.__bind, v, datatype)

        self.call_callback(self.__bind, v)


    def call_callback(self, *args):

        # invoke the callback handler
        callback = self.get_prop("callback")
        if (callback):
            # we get a much better startup performance with the idle handler
            gobject.idle_add(self.__caller, callback, *args)



    def get_widgets(self): raise NotImplementedError

    def _set_enabled(self, value): raise NotImplementedError
    def _set_label(self, value): raise NotImplementedError


    def _setp_bind(self, key, value):

        self.__bind = value


    def _setp_enabled(self, key, value):

        self._setp(key, value)
        self._set_enabled(value)


    def _setp_label(self, key, value):

        self._setp(key, value)
        self._set_label(value)



from main import HOME, REGISTRY_PATH
from plugin.Interface import Interface
from utils.MetaDataRegistry import MetaDataRegistry
from utils import vfs

import commands
import os

_REGFILE = os.path.join(REGISTRY_PATH, "controls.reg")


class ControlRegistry:
    """
    Class for holding a registry of controls. The registry is saved to a
    'controls.reg' file in order to speed up the startup process, since cached
    information can be used.
    """


    def __init__(self, repos):

        self.__registry = {}

        self.__mdreg = MetaDataRegistry(_REGFILE, repos,
                                        self.__find_controls,
                                        self.__register_control)

        self.__search_for_controls()



    #
    # Returns whether the given path has a control.
    #
    def __is_control(self, path):

        # is this test sufficient?
        return os.path.exists(os.path.join(path, "__init__.py"))



    #
    # Returns a list of all controls in the given repository.
    #
    def __find_controls(self, path):

        out = []
        if (not os.path.isdir(path)): return out

        files = os.listdir(path)
        for f in files:
            fpath = os.path.join(path, f)
            if (os.path.isdir(fpath)):
                if (self.__is_control(fpath)): out.append(fpath)
                else:
                    out += self.__find_controls(fpath)

        return out


    #
    # Registers the given control. Registering a control requires us to
    # actually load the control, so this is a critical phase where we have to
    # be careful about malware.
    #
    def __register_control(self, ctrl):

        items = []

        # it's more secure to use a separate process here
        cmd = os.path.join(HOME, "ctrlinfo")
        fail, out = commands.getstatusoutput("%s %s" % (cmd, ctrl))

        if (fail):
            log("Warning: \"%s\" is an invalid control." % ctrl)
            return None

        else:
            log("Registering new control \"%s\"." % ctrl)

        # cut off initial crap; you never know what controls might do during
        # initialization...
        out = out[out.find("PATH:"):]

        for line in out.splitlines():
            if (":" in line):
                key, value = line.split(":")
                value = vfs.unescape_path(value)
            else:
                key = value = ""

            if (not value in self.__registry):
                self.__registry[value] = []

            items.append((key, value))

        return items


    #
    # Searches for new controls.
    #
    def __search_for_controls(self):

        self.__mdreg.scan_repositories()

        self.__registry = {}

        # for all items
        for item in self.__mdreg.get_item_list():
            # go through all key-value pairs
            for k, v in self.__mdreg.get_item(item):
                # being only interested in the IDs
                if (k in ("ID1", "ID2")):
                    # register every single ID
                    for iface in v.split(" "):
                        if (not iface in self.__registry):
                            self.__registry[iface] = []
                        self.__registry[iface].append(item)



    #
    # Returns a control matching the given interface.
    #
    def get_control(self, iface, second_try = False):

        candidates = self.__registry.get(iface, [])
        for i in candidates:
            if (os.path.isdir(i)): return i

        if (not second_try):
            # search for new controls and give it a second try
            self.__search_for_controls()
            return self.get_control(iface, True)

        else:
            return ""



from Vault import Vault


class ElementWrapper(object):
    """
      The ElementWrapper wraps Element objects for use in the sandbox.
      By wrapping it up, there remains no way to access the internals of the
      element. Only the properties and authorized methods will be available.
    """

    def __init__(self, target):

        self.__dict__["_ElementWrapper__target"] = Vault(target)


    def __setattr__(self, name, value):

        self.__target(open).set_prop(name, value)


    def __getattr__(self, name):

        if name in self.__target(open).AUTHORIZED_METHODS:
            return getattr(self.__target(open), name)

        return self.__target(open).get_prop(name)



from TargetCanvas import TargetCanvas
from utils.datatypes import *


#
# Class for plotting graphs.
#
class TargetPlotter(TargetCanvas):

    def __init__(self, name, parent):

        # history of plot values
        self.__history = []

        # colors
        self.__bgcolor = ""
        self.__color = "navy"

        # scale range, initial value
        self.__minscale = 0
        self.__maxscale = 100

        TargetCanvas.__init__(self, name, parent)

        self._register_property("value", TYPE_FLOAT,
                                self._setp_value, self._getp)
        self._register_property("bg-color", TYPE_STRING,
                                self._setp_bgcolor, self._getp)
        self._register_property("color", TYPE_STRING,
                                self._setp_color, self._getp)
        self._register_property("size", TYPE_INT,
                                self._setp, self._getp)
        self._setp("size", 50)
        self._register_property("scala", TYPE_INT,
                                self._setp, self._getp)
        self._setp("scala", 0)
        self._register_property("scala-color", TYPE_STRING,
                                self._setp, self._getp)
        self._setp("scala-color", "black")
        self._register_property("scala-font", TYPE_STRING,
                                self._setp, self._getp)
        self._setp("scala-font", "Sans")

        # scale-bidir: not only scale the upper bound, but also the lower one
        self._register_property("scale-bidir", TYPE_BOOL,
                                self._setp, self._getp)
        self._setp("scale-bidir", False)
        # scale-holdmax: keep the maximum 'all-time' scaling, not only
        # max(history)
        self._register_property("scale-holdmax", TYPE_BOOL,
                                self._setp, self._getp)
        self._setp("scale-holdmax", False)


    #
    # Plots the values from the history.
    #
    def __plot(self):

        bg, fg = self.__bgcolor, self.__color
        min_value = self.__minscale
        max_value = self.__maxscale
        size = self.get_prop("size")

        x = 0.0
        delta_x = 100.0 / (size - 1)

        if (bg):
            try:
                color, opacity = self.__parse_color(bg)
            except ValueError, exc:
                log(`exc`)
                return

            body = "<rect x=\"0\" y=\"0\" width=\"100\" height=\"100\" " \
                   "style=\"stroke:%s;stroke-opacity:%d%%;fill:%s;" \
                   "fill-opacity:%d%%\"/>" % (color, opacity, color, opacity)
        else:
            body = ""

        if (self.__history):
            if ((self.get_prop("scale-holdmax")) == False):
                # reset scale range to default:
                max_value = 100
                min_value = 0
            max_value = max(max_value, max(self.__history))
            if (self.get_prop("scale-bidir")):
                min_value = min(min_value, min(self.__history))
            self.__minscale = min_value
            self.__maxscale = max_value
            scale = 100.0 / (max_value - min_value)

            x, y = x, self.__history[0]
            body += "<path d=\"M%f %f " % (x, 100 - (y - min_value) * scale)
            x += delta_x

            for y in self.__history[1:]:
                y = 100 - (y - min_value) * scale
                body += "L%(x)f %(y)f " % vars()
                x += delta_x

            try:
                color, opacity = self.__parse_color(fg)
            except ValueError, exc:
                log(`exc`)
                return

            body += "\" style=\"stroke:%s;opacity:%d%%;fill:none\"/>" % \
                     (color, opacity)

        # draw scala
        scala = self.get_prop("scala")
        if (scala):
            scala_color = self.get_prop("scala-color")
            scala_font = self.get_prop("scala-font")
            for i in range(0, max_value + 1, scala):
                body += "<text x=\"0\" y=\"%d\" font-family=\"%s\" " \
                        "font-size=\"%d\" fill=\"%s\">%d</text>" \
                        % (max_value - i, scala_font, 5, scala_color, i)

        gfx = "<svg>" + body + "</svg>"
        self.set_prop("graphics", gfx)



    def __parse_color(self, value):

        if ("#" in value and len(value) == 9):
            opacity = float(int(value[7:], 16) / 255.0) * 100
            return (value[:7], opacity)
        elif (value.isalnum()):
            return (value, 100)
        else:
            raise ValueError("%s is an invalid color." % `value`)



    def _setp_value(self, key, value):

        self.__history.append(value)
        self.__plot()
        if (len(self.__history) > self.get_prop("size")):
            self.__history.pop(0)
        self._setp(key, value)



    def _setp_bgcolor(self, key, value):

        self.__bgcolor = value
        self.__plot()
        self._setp(key, value)


    def _setp_color(self, key, value):

        self.__color = value
        self.__plot()
        self._setp(key, value)




from libdesklets.controls import Interface, Permission

class IHDDTemp(Interface):

    # properties along with their permissions
    available_devices = Permission.READ
    poll_all = Permission.READ
    
    device = Permission.READWRITE
    poll = Permission.READ



#
# Module for telling the user about important stuff after transition to another
# version.
#
# ~/.gdesklets/registry/version keeps the gDesklets version
#
# If it doesn't contain our version, we might want to tell the user something.
#

from main import USERHOME, VERSION
from utils import dialog

import os


#
# Compares the current gDesklets version with the version which previously ran
# and returns True if the versions match.
#
def _check_version():

    path = os.path.join(USERHOME, "registry", "version")
    try:
        version = open(path).read()
    except:
        version = "old :)"

    try:
        open(path, "w").write(VERSION)
    except:
        pass


    return (VERSION == version)




if (not _check_version()):
    if (VERSION.endswith("CVS")):
        # CVS warning :)
        msg = _("This is a CVS version of gDesklets. CVS versions represent " \
                "the current state of development and might be unstable or " \
                "cause problems from time to time.\n\n" \
                "If you're new to gDesklets, it is thus highly recommended to "
                "<b>not</b> use a CVS version, but a stable release!\n\n"
                "If you still want to run the CVS version instead of a "
                "stable release version, we'd highly appreciate it if you "
                "report any weird behavior to the developers.\n\n"
                "Experienced users are encouraged to try a CVS version, "
                "though!")

    elif ("rc" in VERSION):
        msg = _("This is a <b>release candidate</b> of an upcoming gDesklets "
                "release.\nPlease test it and report bugs to "
                "<i>http://bugzilla.gnome.org</i>\n"
                "This version might break your configuration or it won't "
                "restore it from an earlier release. In most cases this is "
                "intentional, since gDesklets is still in its early stages of "
                "development.\n\nThanks for testing this release candidate!")

    else:
        # our informative transition message which we can change with
        # every release
        msg = _("This version of gDesklets does no longer use the GConf "
                "database. If you don't plan to use an older version of "
                "gDesklets, you can safely remove the '/apps/gdesklets' "
                "folder from your GConf database.\n\n"
                "Due to the complete transition of the configuration data "
                "into the '~/.gdesklets' folder, gDesklets 0.35 will not "
                "make use of your previous configuration.\n\n"
                "Sorry if this should cause any inconveniences!")


    # show the dialog
    dialog.info(
        _("Please note (this message will appear only once):"),
        msg)



import Unit
from HashPQueue import HashPQueue


class LayoutObject(object):
    """
    gDesklets Geometry Engine
    (c) 2003 - 2005 Martin Grimme  <martin@gdesklets.org>

    This engine is licensed under the terms of the GNU LGPL.
    

    This class represents a layout object for handling geometry calculations
    for visual elements. It is based on the highly flexible and sophisticated
    geometry engine of gDesklets 0.34.3, but is completely decoupled from the
    rendering backend now and optimized even further.
    """

    __slots__ = ("__margin_right", "__margin_bottom", "__margin_ids",
                 "__enabled",
                 "__children", "__parent", "__relative", "__is_relative",
                 "__relatives", "__anchor",
                 "__real_geometry", "__old_real_geometry", "__geometry",
                 "__border_width", "__bbox",
                 "__geometry_callback", "__action_callback")



    def __init__(self, parent = None):
        """
        Creates a new layout object. You normally only call this for the root
        object of the hierarchy of layout objects, without specifying any parent
        object.
        """
        

        # right margin
        self.__margin_right = HashPQueue(HashPQueue.GREATER_THAN)

        # bottom margin
        self.__margin_bottom = HashPQueue(HashPQueue.GREATER_THAN)

        # the IDs of the margin entries
        self.__margin_ids = (-1, -1)

        # whether the element is enabled
        self.__enabled = True

        # the children of a layout object
        self.__children = []

        # the parent object
        self.__parent = parent

        # the object to which this object is placed relative to
        self.__relative = None
        self.__is_relative = (False, False)
        self.__relatives = []

        # the position of the positioning anchor given in percents of the
        # object's size
        self.__anchor = (0, 0)

        # the layout object's geometry
        self.__real_geometry = (Unit.ZERO, Unit.ZERO, Unit.ZERO, Unit.ZERO)
        self.__old_real_geometry = (Unit.ZERO, Unit.ZERO, Unit.ZERO, Unit.ZERO)
        self.__geometry = (Unit.Unit(), Unit.Unit(), Unit.Unit(), Unit.Unit())

        # the widths of the border
        self.__border_width = (Unit.ZERO, Unit.ZERO, Unit.ZERO, Unit.ZERO)

        # the children bounding box
        self.__bbox = (Unit.ZERO, Unit.ZERO, Unit.ZERO, Unit.ZERO)

        # the geometry callback handler
        self.__geometry_callback = None

        # the action callback handler
        self.__action_callback = None
        


    def new_child(self):
        """
        Creates and returns a new layout object as a child of this layout
        object. Use this method to add new objects.
        """

        child = LayoutObject(self)
        self.__children.append(child)

        return child



    def remove_child(self, child):
        """
        Removes the given child layout object. Raises a ValueError if the object
        does not exist.
        """

        self.__children.remove(child)
        rx, ry, rw, rh = child.get_real_geometry()
        self.__margin_right.remove(rx + rw)
        self.__margin_bottom.remove(ry + rh)
        



    def set_callback(self, handler):
        """
        Sets the callback handler for geometry changes. The callback handler
        will be invoked whenever the geometry of the object changes.

        The handler has the signature:
          handler(src, x, y, width, height)

        The callback notifies you about geometry changes of an object.
        """

        self.__geometry_callback = handler



    def set_action_callback(self, handler):

        self.__action_callback = handler



    def set_enabled(self, value):
        """
        Enables or disables the layout object. Disabled objects don't contribute
        to the layout and act as if they did not exist.
        """

        self.__enabled = value
        self._update_all()


    def set_geometry(self, x = None, y = None, width = None, height = None):
        """
        Sets the geometry values. The values have to be valid Unit objects.
        """

        gx, gy, gw, gh = self.__geometry
        
        if (x): gx = x
        if (y): gy = y
        if (width): gw = width
        if (height): gh = height

        self.__geometry = (gx, gy, gw, gh)
        self._update_all()



    def get_geometry(self):
        """
        Returns the user-given geometry in Unit values.
        """

        return self.__geometry



    def get_real_geometry(self):
        """
        Returns the real geometry of the object, with anchors, and relative and
        percentual positioning resolved.
        """

        #x, y, w, h = self.__real_geometry
        #print "COMPUTE LAYOUT", x, y, w, h
        #if (self.__parent):
        #    parent_width, parent_height = self.__get_parent_size()
            #w1, h1, w2, h2 = self.__parent.get_border_width()
            #print parent_width, w1, w2
            #parent_width = parent_width - w1 - w2
            #parent_height = parent_height - h1 - h2

        #    x.set_100_percent(parent_width.as_px())
        #    y.set_100_percent(parent_height.as_px())
        #    w.set_100_percent(parent_width.as_px())
        #    h.set_100_percent(parent_height.as_px())

        return self.__real_geometry


    def _update_bbox(self, right_id, bottom_id, x, y, w, h, ignore_x, ignore_y):
        """
        Updates the bounding box with the given values.
        """

        self.__margin_right.remove(right_id)
        self.__margin_bottom.remove(bottom_id)

        right_id, bottom_id = -1, -1
        if (not ignore_x):
            right_id = self.__margin_right.insert(x.copy() + w.copy())
        if (not ignore_y):
            bottom_id = self.__margin_bottom.insert(y.copy() + h.copy())

        return (right_id, bottom_id)



    def __get_children_bbox(self):
        """
        Returns the bounding box of the child objects.
        """

        bx1 = Unit.ZERO
        by1 = Unit.ZERO
        bx2 = self.__margin_right.top() or Unit.ZERO
        by2 = self.__margin_bottom.top() or Unit.ZERO
        #print "BBOX", (bx1, by1, bx2, by2), self
        return (bx1, by1, bx2, by2)
        


    def __get_parent_size(self):
        """
        Returns the size of the parent widget.
        """

        parent_width, parent_height = \
                      self.__parent.get_real_geometry()[2:]

        #print "GET P SIZE", parent_width, parent_height
        # don't allow dangerous values for the parent's size
        parent_width = max(Unit.ONE, parent_width)
        parent_height = max(Unit.ONE, parent_height)
        
        return (parent_width, parent_height)


    def set_anchor(self, x, y):
        """
        Sets the positioning anchor to the given place. The coordinates are
        given in percentages of the object's size with values between 0.0
        and 1.0.
        """

        self.__anchor = (x, y)
        self._update_all()
        


    def get_anchor(self):
        """
        Returns the position of the anchor.
        """

        return self.__anchor


    def __is_percentual(self):
        """
        Returns whether the object has some percentual values.
        """

        x, y, w, h = self.__geometry
        return Unit.UNIT_PERCENT in (x.get_unit(), y.get_unit(),
                                     w.get_unit(), h.get_unit())



    def __resolve_anchor(self, x, y, w, h):
        """
        Resolves the anchored position and returns the position of the top-left
        corner of the object.
        """

        ax, ay = self.get_anchor()
        ax = w * ax
        ay = h * ay
        
        return (x - ax, y - ay)





    def _update_all(self):
        """
        Updates the geometry of the layout object and adjusts depending objects
        as well.
        """

        if (self._update_geometry()):
            self._update_relatives()
            self._update_children()
            self._update_parent()

            # objects with percentual geometry sometimes need special treatment
            if (self.__is_percentual()):
                if (self._update_geometry()):
                    self._update_relatives()

                    #self._update_parent()
            #self._update_geometry()
        else:
            self._update_children()


    def _update_geometry(self):
        """
        Updates the geometry of the layout object.
        This does not include notifying the parent object or relative objects
        about the change.

        Returns True if something changed, False otherwise.
        """
        
        x, y, w, h = self.__real_geometry
        ux, uy, uw, uh = self.__geometry

        # adjust percentage stuff
        if (self.__parent):
            parent_width, parent_height = self.__get_parent_size()
            w1, h1, w2, h2 = self.__parent.get_border_width()
            parent_width = parent_width - w1 - w2
            parent_height = parent_height - h1 - h2

            if (parent_width.is_unset()): parent_width = w
            if (parent_height.is_unset()): parent_height = h
            
            ux.set_100_percent(parent_width.as_px())
            uw.set_100_percent(parent_width.as_px())
            uy.set_100_percent(parent_height.as_px())
            uh.set_100_percent(parent_height.as_px())

        # compute geometry
        if (ux.is_unset()): new_x = x
        else: new_x = ux

        if (uy.is_unset()): new_y = y
        else: new_y = uy

        if (self.__enabled):
            if (uw.is_unset()): new_w = self.__get_value_for_unset(width = w)
            else: new_w = uw

            if (uh.is_unset()): new_h = self.__get_value_for_unset(height = h)
            else: new_h = uh

        else:
            new_w = Unit.ZERO
            new_h = Unit.ZERO

        # handle relative positioning
        if (self.__relative):
            is_rel_x, is_rel_y = self.__is_relative

            # get the coords of the anchor of the relative
            rx, ry, rw, rh = self.__relative.get_real_geometry()
            anchor_x, anchor_y = self.__relative.get_anchor()
            anchor_x = rw * anchor_x
            anchor_y = rh * anchor_y

            # get the remaining amount of pixels from the anchor to the
            # bottom/right edge of the relative
            tw = rw - anchor_x
            th = rh - anchor_y

            if (not ux.is_unset()): new_x += rx + anchor_x
            else: new_x = rx + anchor_x
            if (not uy.is_unset()): new_y += ry + anchor_y
            else: new_y = ry + anchor_y
            if (is_rel_x): new_x += tw
            if (is_rel_y): new_y += th
            #print "REL", new_x, new_y, new_w, new_h
        #end if
        

        # get coords of the top left corner, i.e. resolve the anchor
        new_x, new_y = self.__resolve_anchor(new_x, new_y, new_w, new_h)

        old_x, old_y, old_w, old_h = self.__old_real_geometry
        if ((new_x, new_y, new_w, new_h) != self.__old_real_geometry):
            self.__real_geometry = (new_x.copy(), new_y.copy(),
                                    new_w.copy(), new_h.copy())
            self.__old_real_geometry = (new_x.copy(), new_y.copy(),
                                        new_w.copy(), new_h.copy())

            # update the parent's children bounding box
            if (self.__parent):
                right_id, bottom_id = self.__margin_ids
                ignore_x = (Unit.UNIT_PERCENT in
                            (ux.get_unit(), uw.get_unit()))
                ignore_y = (Unit.UNIT_PERCENT in
                            (uy.get_unit(), uh.get_unit()))
                self.__margin_ids = \
                        self.__parent._update_bbox(right_id, bottom_id,
                                                   new_x, new_y, new_w, new_h,
                                                   ignore_x, ignore_y)

            # call the geometry callback
            if (self.__geometry_callback):
                self.__geometry_callback(self, new_x, new_y, new_w, new_h)

            return True

        else:
            return False



    def _update_relatives(self):
        """
        Updates the geometry of the objects placed relative to this object.
        """

        for r in self.__relatives:
            r._update_geometry()
            r._update_relatives()


    def _update_children(self):
        """
        Updates the geometry of the child objects of this object. This only
        affects children with percentual geometry values.
        """

        if (not self.__enabled): return
        x, y, w, h = self.get_geometry()
        for c in self.__children:
            ux, uy, uw, uh = c.get_geometry()
            if (Unit.UNIT_PERCENT in (ux.get_unit(), uy.get_unit(),
                                      uw.get_unit(), uh.get_unit())):
                c._update_geometry()
                c._update_relatives()
                c._update_children()
        #end for



    def _update_parent(self):
        """
        Updates the geometry of the parent object. This only affects the parent
        if it does not have a fixed size.
        """

        if (self.__parent):
            x, y, w, h = self.__parent.get_geometry()
            if (Unit.Unit() in (w, h)):
                self.__parent._update_all()
    


    def set_relative_to(self, other, is_rel_x, is_rel_y):
        """
        Places the object relative to another one.
        """

        self.__relative = other
        self.__is_relative = (is_rel_x, is_rel_y)
        other._add_relative(self)
        self._update_all()



    def _add_relative(self, r):
        """
        Adds a new relative object.
        """

        self.__relatives.append(r)


    def _remove_relative(self, r):
        """
        Removes a relative object.
        """

        try:
            self.__relatives.remove(r)
        except:
            pass



    def set_border_width(self, left, top, right, bottom):
        """
        Sets the widths of the four borders of the layout object.
        A border-width > 0 results in inner padding.
        """

        self.__border_width = (left, top, right, bottom)
        self._update_all()



    def get_border_width(self):
        """
        Returns the widths of the four borders as a 4-tuple.
        """

        return self.__border_width



    def __get_value_for_unset(self, x = None, y = None,
                             width = None, height = None):
        """
        Returns the geometry value that should be used for unset values.
        """

        bx1, by1, bx2, by2 = self.__get_children_bbox()
        bw1, bh1, bw2, bh2 = self.__border_width

        if (x): return x
        elif (y): return y
        elif (width): return bx2 + bw1 + bw2
        elif (height): return by2 + bh1 + bh2



    def send_action(self, x, y, *args):

        if (self.__action_callback):
            self.__action_callback(self, x, y, *args)

        bw, bh, nil, nil = self.__border_width
        x -= bw
        y -= bh

        for c in self.__children:
            cx, cy, cw, ch = c.get_real_geometry()
            if (cx <= x < cx + cw and cy <= y < cy + ch):
                c.send_action(x - cx, y - cy, *args)
        #end for


    def dump(self, indent = 0):
        """
        For debugging:
        Dumps the geometry values of the layout object together with all
        child-objects to stdout.
        """

        print " " * indent,
        print self.get_geometry(), "---", \
              map(lambda f: f.as_px(), self.get_real_geometry())
        for c in self.__children:
            c.dump(indent + 2)



if (__name__ == "__main__"):

    help(LayoutObject)

    def on_action(src, x, y, arg):

        print x, y, arg
        

    def on_change(src, x, y, w, h):

        print src
        print "  ", x, y, w, h


    r = LayoutObject()
    r.set_geometry(Unit.ZERO, Unit.ZERO,
                   Unit.Unit(1280, Unit.UNIT_PT), Unit.Unit(1024, Unit.UNIT_PT))
    a = r.new_child()

    a.set_geometry(Unit.Unit(50, Unit.UNIT_PERCENT),
                   Unit.Unit(10, Unit.UNIT_PERCENT),
                   Unit.Unit(), Unit.Unit())
    a.set_border_width(Unit.Unit(10, Unit.UNIT_PX),
                       Unit.Unit(10, Unit.UNIT_PX),
                       Unit.Unit(10, Unit.UNIT_PX),
                       Unit.Unit(10, Unit.UNIT_PX))

    for x in range(0, 7):
        for y in range(0, 5):
            new = a.new_child()
            new.set_callback(on_change)
            new.set_action_callback(on_action)
            new.set_geometry(Unit.Unit(x, Unit.UNIT_PT),
                             Unit.Unit(y, Unit.UNIT_PT),
                             Unit.Unit(50, Unit.UNIT_PT),
                             Unit.Unit(10, Unit.UNIT_PT))
            new.set_anchor(1.0, 0.2)
        
        
    print a.get_real_geometry()
    a.send_action(Unit.Unit(5, Unit.UNIT_PT),
                  Unit.Unit(4, Unit.UNIT_PT), "test")



from main import CREDITS

import gobject
import gtk
import os


class AboutDialog(gtk.Window):

    """ Class for a fancy "About" dialog. """

    def __init__(self, path):

        self.__is_stopped = True
        self.__scroller_values = ()

        gtk.Window.__init__(self, gtk.WINDOW_TOPLEVEL)
        self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_DIALOG)
        self.set_position(gtk.WIN_POS_CENTER)
        self.set_resizable(False)
        self.set_title(_("About gDesklets"))
        self.connect("button-press-event", self.__on_close)
        self.connect("key-press-event", self.__on_close)
        self.connect("delete-event", self.__on_close)
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK)

        fixed = gtk.Fixed()
        self.add(fixed)

        img = gtk.Image()
        img.set_from_file(os.path.join(path, "about.png"))
        fixed.put(img, 0, 0)

        width, height = img.size_request()

        scroller = gtk.Fixed()
        scroller.set_size_request(width, height - 32)
        fixed.put(scroller, 0, 32)

        text = ""
        for header, body in CREDITS:
            text += "<b>" + header + "</b>\n\n"
            text += "\n".join(body)
            text += "\n\n\n\n"
        text = "<big>" + text.strip() + "</big>"

        credits = gtk.Label(text)
        credits.set_use_markup(True)
        credits.set_justify(gtk.JUSTIFY_CENTER)

        lbl_width, lbl_height = credits.size_request()
        scroller.put(credits, (width - lbl_width) / 2, height)

        self.__scroller = scroller
        self.__credits = credits

        self.__scroller_values = (height - 32, -lbl_height,
                                  (width - lbl_width) / 2)


    def __scroll(self, ycoord):

        begin, end, xcoord = self.__scroller_values
        self.__scroller.move(self.__credits, xcoord, ycoord)
        ycoord -= 1

        if (ycoord < end):
            ycoord = begin

        if (not self.__is_stopped):
            gobject.timeout_add(50, self.__scroll, ycoord)

        return False


    def __on_close(self, *args):

        self.__is_stopped = True
        self.hide()

        return True


    def show(self):

        if (not self.__is_stopped):
            return

        self.show_all()

        self.__is_stopped = False
        begin, end, xcoord = self.__scroller_values
        gobject.timeout_add(0, self.__scroll, begin)




from Sensor import Sensor

import threading


class MetaSensor(Sensor):

	"""
	A MetaSensor is a specialisation of a Sensor, that has efficient
	threaded behaviour built-in to it. Basically, your sensor provides
	3 things for the metasensor :

	- A fetcher method that fetchs data from some source
	- A displayer method that displays this data
	- A timer method that pauses between updates
	
	Once, you initialise a MetaSensor with these three parts,
	it will ensure that the datasource is polled at regular intervals,
	and that the display code is only called if there is something new
	to display.
	"""
	

	def __init__(self, retriever, rtimer, displayer):

		"""
		@retriever : function that retrieves data
		@type retriever: callable, returns True when new data
		available.

		@rtimer : function that return a sleep duration
		@type rtimer : callable. Returns int > 0.

		@displayer : function that displays data
		@type displayer : callable


		On config changed, only @displayer is called.
		@retriever lives is life, polling data and calling
		@displayer when needed.

		That's the best we can do with these old deprecated Sensor.
		
		"""
		
		Sensor.__init__(self)

		self.__retriever = retriever
		self.__rtimer    = rtimer
		self.__displayer = displayer

		self.__retriever_event = threading.Event()
		self.__displayer_event = threading.Event()
		
		self.watch_stop( self.__on_stop )
		self.watch_config( self.__on_update )


		def once():

			self.add_thread( self.__retriever_thread )
			self.add_thread( self.__displayer_thread )
			self._start = Null

		self._start = once



	def watch_stop(self, callback):
		"""
		You can't override me :P
		"""
		pass


	def watch_config(self, callback):
		"""
		You can't override me :P
		"""
		pass


	def __on_stop(self):

		self.__retriever_event.set()
		self.__displayer_event.set()



	def __on_update(self, *dummy):

		self.__displayer_event.set()



	def __retriever_thread(self):

		while not self.is_stopped():

			if self.__retriever():
				self.__displayer_event.set()

			self.__retriever_event.wait( self.__rtimer() )




	def __displayer_thread(self):

		while not self.is_stopped():

			self.__displayer_event.wait()
			self.__displayer()
			self.__displayer_event.clear()
			



from factory.SensorFactory import SensorFactory
from main.Control import Control
from ISensor import ISensor

import gtk


#
# This control wraps legacy sensors to make them still usable.
# Sensors are deprecated and shouldn't be used in new stuff. This control is
# solely meant for retaining backwards compatibility.
#
class Sensor(Control, ISensor):

    def __init__(self):

        self.__sensor = None
        self.__sensor_factory = SensorFactory()

        self.__output = None
        self.__menu = None


        Control.__init__(self)


    #
    # Loads the given sensor with arguments.
    #
    def __set_sensor(self, value):

        module, args = value
        sensor = self.__sensor_factory.create_sensor(module, args)
        if (sensor):
            sensor.add_observer(self.__on_observe_sensor)
        else:
            raise RuntimeError("Could not load sensor")

        self.__sensor = sensor


    #
    # Sends an action to the sensor.
    #
    def __set_action(self, value):

        callname, path, allargs = value
        self.__sensor.send_action(callname, path, allargs)

    def __set_config_id(self, value): self.__sensor.set_config_id(value)

    def __set_stop(self, value): self.__sensor.stop()

    def __get_output(self): return self.__output
    def __get_menu(self): return self.__menu
    def __get_configurator(self): return self.__sensor.get_configurator()


    #
    # Observer for the sensor.
    #
    def __on_observe_sensor(self, src, cmd, data):

        # propagate the incoming sensor output
        if (cmd == src.OBS_OUTPUT):
            self.__output = data
            self._update("output")

        elif (cmd == src.OBS_CMD_MENU):
            self.__menu = data
            self._update("menu")


    sensor = property(None, __set_sensor, doc = "the sensor")
    action = property(None, __set_action, doc = "the action to perform")
    config_id = property(None, __set_config_id, doc = "the config ID")
    stop = property(None, __set_stop, doc = "stops the sensor")
    output = property(__get_output, None,
                      doc = "the output data of the sensor")
    menu = property(__get_menu, None,
                      doc = "the menu data of the sensor")
    configurator = property(__get_configurator, None,
                            doc = "the configurator of the sensor")


def get_class(): return Sensor



#! /usr/bin/env python
# A simple script to test controls interactively.

from plugin.Interface import Interface
from main import HOME
from plugin import Permission


import sys
import os
import code
import __builtin__

if "." not in sys.path: sys.path.append(".")
if HOME not in sys.path: sys.path.append(HOME)


try:
    path = os.path.abspath(sys.argv[1])
    folder, base = os.path.split(path)
except:
    sys.exit("Usage: test-control.py <control-directory>")
    

cwd = os.getcwd()
os.chdir(folder)
try:
    module = __import__(base)
    os.chdir(base)
    clss = module.get_class()
    ctrl = clss()
    
except IOError:
    sys.exit("Could not load control %s." % (path))

print
print Interface.text_describe(clss)

__builtin__.ctrl = ctrl
code.InteractiveConsole().interact("Use 'ctrl' to access the control. "
                                   "Press Ctrl+D to quit.")



from libdesklets.controls import Control
from IEventPipe import IEventPipe


#
# Control for piping events to other desklets.
#
# This control acts as a hub providing a push based notification system with
# different channels of interests.
#
# You can use this control to subscribe to channels in order to automatically
# receive messages:
#
#   epipe.subscriptions = ["SideCandy", "System"]
#
# By binding a handler to the "event" property, you can receive broadcasted
# messages:
#
#   epipe.bind("event", on_epipe_message)
#
# In order to broadcast a message to other desklets, you put the message into
# the "event" property:
#
#   epipe.event = ("SideCandy", "my message")
#
# A message is a list of elements, where the first element is the name of the
# channel, and the second is the name of the message. The list may also contain
# further, arbitrary elements, but the channel and the message name must be
# strings.
#
class EventPipe(Control, IEventPipe):

    # table: channel -> [subscribers]
    __subscribers = {}
    
    

    def __init__(self):

        self.__my_subscriptions = []
        self.__event = None
        
        
        Control.__init__(self)


    def __get_subs(self):

        return self.__my_subscriptions


    def __set_subs(self, subs):

        # remove from subscribers list
        for channel in self.__my_subscriptions:
            self.__subscribers[channel].remove(self)

        # add to subscribers list
        for channel in subs:
            if (not channel in self.__subscribers):
                self.__subscribers[channel] = []
            if (not self in self.__subscribers[channel]):
                self.__subscribers[channel].append(self)
                
        self.__my_subscriptions = subs


    def __get_event(self):

        return self.__event
    

    def __set_event(self, message):

        self.__broadcast(message)


    def __broadcast(self, message):

        channel = message[0]
        for s in self.__subscribers.get(channel, []):
            s._send_message(message)


    def _send_message(self, message):

        self.__event = message
        self._update("event")
        self.__event = None


    def _shutdown(self):

        # remove from subscribers list
        for channel in self.__my_subscriptions:
            self.__subscribers[channel].remove(self)


    subscriptions = property(__get_subs, __set_subs,
                             doc = "The subscribed channels")
    event = property(__get_event, __set_event, "The event message; set to send "
                                             "events, bind() to receive events")



def get_class(): return EventPipe



from libdesklets.system.Arch import Arch

import re
import struct


class Generic(Arch):

    def __init__(self):

        Arch.__init__(self)

        self.__net_devices = re.compile('^\s*(\w+):.*mtu', re.M
                                        ).findall( os.popen('/sbin/ifconfig -a').read() )

        # cpu0 at mainbus0: MIPS R3000A CPU (0x230) Rev. 3.0 with MIPS R3010 FPC Rev. 3.0
        # cpu0: 64KB/4B direct-mapped Instruction cache, 64 TLB entries
        # cpu0: 64KB/4B direct-mapped write-through Data cache

        m = re.compile('^cpu0 at mainbus0: (.*?) CPU', re.M
                       ).search(open("/var/run/dmesg.log").read() )

        self.__model = m.group(1)



    def net_devices(self):
        """
        @return : all available network devices
        @rtype  : list
        """

        return (self.__net_devices)



    def cpu_model(self):
        """
        @return : model/type of installed processor
        @rtype  : str
        """

        return (self.__model)



    def users(self):
        """
        @return : number of connected users
        @rtype  : int
        """

        # man utmp
        # don't know if
        # sizeof(struct utmp) == 36
        # offsetof(struct utmp, ut_name) == 8
        # on every NetBSD arch
        # mips3000 : ok

        count = 0
        data = open('/var/run/utmp', 'rb').read()

        for i in range(8, len(data), 36):

            ut_name = struct.unpack('c', data[i:i+1])[0]

            if (ut_name != '\0'):
                count += 1

        return count
